import json

from langchain.tools import BaseTool

from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input


class SolanaCybersCreateCoinTool(BaseTool):
    name: str = "cybers_create_coin"
    description: str = """
    Creates a new coin using the CybersManager.

    Input: A JSON string with:
    {
        "name": "string, the name of the coin",
        "symbol": "string, the symbol of the coin",
        "image_path": "string, the file path to the coin's image",
        "tweet_author_id": "string, the Twitter ID of the coin's author",
        "tweet_author_username": "string, the Twitter username of the coin's author"
    }

    Output:
    {
        "coin_id": "string, the unique ID of the created coin",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "name": {"type": str, "required": True},
                "symbol": {"type": str, "required": True},
                "image_path": {"type": str, "required": True},
                "tweet_author_id": {"type": str, "required": True},
                "tweet_author_username": {"type": str, "required": True}
            }
            validate_input(data, schema)

            name = data["name"]
            symbol = data["symbol"]
            image_path = data["image_path"]
            tweet_author_id = data["tweet_author_id"]
            tweet_author_username = data["tweet_author_username"]

            coin_id = await self.solana_kit.cybers_create_coin(
                name=name,
                symbol=symbol,
                image_path=image_path,
                tweet_author_id=tweet_author_id,
                tweet_author_username=tweet_author_username
            )
            return {
                "coin_id": coin_id,
                "message": "Success"
            }
        except Exception as e:
            return {
                "coin_id": None,
                "message": f"Error creating coin: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError(
            "This tool only supports async execution via _arun. Please use the async interface."
        )




class BN:
    def __init__(self, value: int):
        self.value = int(value)
    
    def to_bytes(self, length: int, byteorder: str, signed: bool = False) -> bytes:
        return self.value.to_bytes(length, byteorder, signed=signed)
    
    def __add__(self, other):
        return BN(self.value + int(other))
    
    def __sub__(self, other):
        return BN(self.value - int(other))
    
    def __mul__(self, other):
        return BN(self.value * int(other))
    
    def __floordiv__(self, other):
        return BN(self.value // int(other))
    
    def __mod__(self, other):
        return BN(self.value % int(other))
    
    def __neg__(self):
        return BN(-self.value)
    
    def __repr__(self):
        return f"BN({self.value})"
    
    def __int__(self):
        return self.value
    
    def is_neg(self):
        return self.value < 0


from agentipy.agent import SolanaAgentKit
from .vault import (
    GetDriftLendBorrowApyTool,
    CreateDriftVaultTool,
    UpdateDriftVaultDelegateTool,
    UpdateDriftVaultTool,
    GetDriftVaultInfoTool,
    DepositIntoDriftVaultTool,
    RequestWithdrawalFromDriftVaultTool,
    WithdrawFromDriftVaultTool,
    DeriveDriftVaultAddressTool,
    TradeUsingDelegatedDriftVaultTool
)
from .account import (
    CreateDriftUserAccountTool,
    DepositToDriftUserAccountTool,
    WithdrawFromDriftUserAccountTool,
    CheckIfDriftAccountExistsTool,
    DriftUserAccountInfoTool
)
from .market import (
    GetAvailableDriftMarketsTool,
    TradeUsingDriftPerpAccountTool,
    GetDriftPerpMarketFundingRateTool,
    GetDriftEntryQuoteOfPerpTradeTool,
    DriftSwapSpotTokenTool
)
from .insurance import (
    StakeToDriftInsuranceFundTool,
    RequestUnstakeFromDriftInsuranceFundTool,
    UnstakeFromDriftInsuranceFundTool
)

def get_drift_tools(solana_kit: SolanaAgentKit):
    
    return [
        # Vault tools
        GetDriftLendBorrowApyTool(solana_kit=solana_kit),
        CreateDriftVaultTool(solana_kit=solana_kit),
        UpdateDriftVaultDelegateTool(solana_kit=solana_kit),
        UpdateDriftVaultTool(solana_kit=solana_kit),
        GetDriftVaultInfoTool(solana_kit=solana_kit),
        DepositIntoDriftVaultTool(solana_kit=solana_kit),
        RequestWithdrawalFromDriftVaultTool(solana_kit=solana_kit),
        WithdrawFromDriftVaultTool(solana_kit=solana_kit),
        DeriveDriftVaultAddressTool(solana_kit=solana_kit),
        TradeUsingDelegatedDriftVaultTool(solana_kit=solana_kit),
        
        # Account tools
        CreateDriftUserAccountTool(solana_kit=solana_kit),
        DepositToDriftUserAccountTool(solana_kit=solana_kit),
        WithdrawFromDriftUserAccountTool(solana_kit=solana_kit),
        CheckIfDriftAccountExistsTool(solana_kit=solana_kit),
        DriftUserAccountInfoTool(solana_kit=solana_kit),
        
        # Market tools
        GetAvailableDriftMarketsTool(solana_kit=solana_kit),
        TradeUsingDriftPerpAccountTool(solana_kit=solana_kit),
        GetDriftPerpMarketFundingRateTool(solana_kit=solana_kit),
        GetDriftEntryQuoteOfPerpTradeTool(solana_kit=solana_kit),
        DriftSwapSpotTokenTool(solana_kit=solana_kit),
        
        # Insurance tools
        StakeToDriftInsuranceFundTool(solana_kit=solana_kit),
        RequestUnstakeFromDriftInsuranceFundTool(solana_kit=solana_kit),
        UnstakeFromDriftInsuranceFundTool(solana_kit=solana_kit)
    ]


from typing import Dict, List, Optional

import nacl.signing
from solana.rpc.api import Client as SolanaClient
from solana.rpc.commitment import Confirmed
from solana.rpc.types import TxOpts
from solders.instruction import Instruction  # type: ignore
from solders.keypair import Keypair  # type: ignore
from solders.pubkey import Pubkey  # type: ignore
from solders.transaction import Transaction  # type: ignore


class SolanaTransaction:
    """Transaction parameters for Solana."""
    def __init__(
        self, 
        instructions: List[Instruction], 
        accounts_to_sign: Optional[List[Keypair]] = None
    ):
        self.instructions = instructions
        self.accounts_to_sign = accounts_to_sign


class SolanaWalletClient:
    """Solana wallet implementation."""

    def __init__(self, client: SolanaClient, keypair: Keypair):
        self.client = client
        self.keypair = keypair

    def get_address(self) -> str:
        return str(self.keypair.pubkey())

    def sign_message(self, message: str) -> Dict[str, str]:
        message_bytes = message.encode("utf-8")
        signed = nacl.signing.SigningKey(self.keypair.secret()).sign(message_bytes)
        return {"signature": signed.signature.hex()}

    def balance_of(self, address: str) -> Dict:
        pubkey = Pubkey.from_string(address)
        balance_lamports = self.client.get_balance(pubkey).value
        return {
            "decimals": 9,
            "symbol": "SOL",
            "value": str(balance_lamports / 10**9),
            "in_base_units": str(balance_lamports),
        }

    def send_transaction(self, transaction: SolanaTransaction) -> Dict[str, str]:
        recent_blockhash = self.client.get_latest_blockhash().value.blockhash
        tx = Transaction()
        tx.recent_blockhash = recent_blockhash
        tx.fee_payer = self.keypair.pubkey()

        for instruction in transaction.instructions:
            tx.add(instruction)

        signers = [self.keypair]
        if transaction.accounts_to_sign:
            signers.extend(transaction.accounts_to_sign)

        tx.sign(*signers)
        result = self.client.send_transaction(
            tx,
            *signers,
            opts=TxOpts(skip_preflight=False, max_retries=10, preflight_commitment=Confirmed),
        )
        self.client.confirm_transaction(result.value, commitment=Confirmed)
        return {"hash": str(result.value)}



from allora_sdk.v2.api_client import (PriceInferenceTimeframe,
                                      PriceInferenceToken, SignatureFormat)
from mcp.types import Tool

from agentipy.tools.use_allora import AlloraManager

ALLORA_ACTIONS = {
    "GET_ALL_TOPICS": Tool(
        name="GET_ALL_TOPICS",
        description="Get all topics from Allora's API. input_schema Example: {}",
        inputSchema={},
        handler=lambda agent, params: AlloraManager(agent).get_all_topics(),
    ),

    "GET_PRICE_PREDICTION": Tool(
        name="GET_PRICE_PREDICTION",
        description=(
            "Fetch a future price prediction for BTC or ETH for a given timeframe from the Allora Network. "
            "input_schema Example: { asset: string, timeframe: string, signature_format: string = 'ETHEREUM_SEPOLIA' }"
        ),
        inputSchema={
            "asset": {
                "type": "string",
                "description": "Crypto asset symbol (BTC or ETH).",
                "enum": ["BTC", "ETH"],
            },
            "timeframe": {
                "type": "string",
                "description": "Prediction timeframe (FIVE_MINUTES or EIGHT_HOURS).",
                "enum": ["FIVE_MINUTES", "EIGHT_HOURS"],
            },
            "signature_format": {
                "type": "string",
                "description": "Blockchain signature format (default: ETHEREUM_SEPOLIA).",
                "enum": ["ETHEREUM_SEPOLIA", "ETHEREUM_MAINNET"],
                "default": "ETHEREUM_SEPOLIA",
            },
        },
        handler=lambda agent, params: AlloraManager(agent).get_price_prediction(
            asset=PriceInferenceToken[params["asset"]],
            timeframe=PriceInferenceTimeframe[params["timeframe"]],
            signature_format=SignatureFormat[params.get("signature_format", "ETHEREUM_SEPOLIA")],
        ),
    ),

    "GET_INFERENCE_BY_TOPIC_ID": Tool(
        name="GET_INFERENCE_BY_TOPIC_ID",
        description="Fetch inference data for a specific topic ID. input_schema Example: { topic_id: number }",
        inputSchema={
            "topic_id": {
                "type": "integer",
                "description": "Topic ID to fetch inference data for."
            },
        },
        handler=lambda agent, params: AlloraManager(agent).get_inference_by_topic_id(params["topic_id"]),
    ),
}



import logging
from typing import Any, Dict, List, Optional

import requests

from agentipy.agent import SolanaAgentKit
from agentipy.utils.agentipy_proxy.utils import encrypt_private_key

logger = logging.getLogger(__name__)

class SolutiofiManager:
    """
    Manager class to handle Solutiofi functionalities such as closing accounts,
    burning tokens, merging tokens, and spreading tokens.
    """

    @staticmethod
    def close_accounts(
        agent: SolanaAgentKit,
        mints: List[str],
    ) -> Optional[Dict[str, Any]]:
        """
        Closes token accounts.

        :param agent: An instance of SolanaAgentKit.
        :param mints: List of mint addresses to close.
        :return: List of transaction signatures or error details.
        """
        try:
            if not mints:
                raise ValueError("Mints list cannot be empty.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "solutiofi_api_key": agent.solutiofi_api_key,
                "mints": mints,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/solutiofi/close-accounts",
                json=payload,
                headers={"Content-Type": "application/json"},
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {"success": True, "transactions": data.get("value"), "message": data.get("message")}
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during closing accounts: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except Exception as error:
            logger.error(f"Unexpected error during closing accounts: {error}", exc_info=True)
            return {"success": False, "error": str(error)}

    @staticmethod
    def burn_tokens(
        agent: SolanaAgentKit,
        mints: List[str],
    ) -> Optional[Dict[str, Any]]:
        """
        Burns tokens from specific mint addresses.

        :param agent: An instance of SolanaAgentKit.
        :param mints: List of mint addresses to burn tokens from.
        :return: List of transaction signatures or error details.
        """
        try:
            if not mints:
                raise ValueError("Mints list cannot be empty.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "solutiofi_api_key": agent.solutiofi_api_key,
                "mints": mints,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/solutiofi/burn-tokens",
                json=payload,
                headers={"Content-Type": "application/json"},
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {"success": True, "transactions": data.get("value"), "message": data.get("message")}
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during burning tokens: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except Exception as error:
            logger.error(f"Unexpected error during burning tokens: {error}", exc_info=True)
            return {"success": False, "error": str(error)}

    @staticmethod
    def merge_tokens(
        agent: SolanaAgentKit,
        input_assets: List[Dict[str, Any]],
        output_mint: str,
        priority_fee: str,
    ) -> Optional[Dict[str, Any]]:
        """
        Merges tokens into a single output mint.

        :param agent: An instance of SolanaAgentKit.
        :param input_assets: List of input assets.
        :param output_mint: The mint address of the output token.
        :param priority_fee: Priority fee level ('fast', 'turbo', 'ultra').
        :return: List of transaction signatures or error details.
        """
        try:
            if not input_assets or not output_mint or not priority_fee:
                raise ValueError("Input assets, output mint, and priority fee are required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "solutiofi_api_key": agent.solutiofi_api_key,
                "inputAssets": input_assets,
                "outputMint": output_mint,
                "priorityFee": priority_fee,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/solutiofi/merge-tokens",
                json=payload,
                headers={"Content-Type": "application/json"},
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {"success": True, "transactions": data.get("value"), "message": data.get("message")}
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during merging tokens: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except Exception as error:
            logger.error(f"Unexpected error during merging tokens: {error}", exc_info=True)
            return {"success": False, "error": str(error)}

    @staticmethod
    def spread_token(
        agent: SolanaAgentKit,
        input_asset: Dict[str, Any],
        target_tokens: List[Dict[str, Any]],
        priority_fee: str,
    ) -> Optional[Dict[str, Any]]:
        """
        Spreads an input token into multiple target tokens.

        :param agent: An instance of SolanaAgentKit.
        :param input_asset: The asset to be spread.
        :param target_tokens: List of target token distributions.
        :param priority_fee: Priority fee level ('fast', 'turbo', 'ultra').
        :return: List of transaction signatures or error details.
        """
        try:
            if not input_asset or not target_tokens or not priority_fee:
                raise ValueError("Input asset, target tokens, and priority fee are required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "solutiofi_api_key": agent.solutiofi_api_key,
                "inputAsset": input_asset,
                "targetTokens": target_tokens,
                "priorityFee": priority_fee,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/solutiofi/spread-token",
                json=payload,
                headers={"Content-Type": "application/json"},
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {"success": True, "transactions": data.get("value"), "message": data.get("message")}
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during spreading tokens: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except Exception as error:
            logger.error(f"Unexpected error during spreading tokens: {error}", exc_info=True)
            return {"success": False, "error": str(error)}



import logging
from typing import Any, Dict, Optional

import requests

from agentipy.agent import SolanaAgentKit
from agentipy.utils.agentipy_proxy.utils import encrypt_private_key

logger = logging.getLogger(__name__)

class AdrenaTradeManager:
    @staticmethod
    def close_perp_trade_short(
        agent: SolanaAgentKit,
        price: float,
        trade_mint: str,
    ) -> Optional[Dict[str, Any]]:
        """
        Closes a perpetual short trade.

        :param agent: An instance of SolanaAgentKit.
        :param price: Trade closing price.
        :param trade_mint: The mint address of the trade asset.
        :return: Transaction signature or error details.
        """
        try:
            if not all([price, trade_mint]):
                raise ValueError("Price and trade_mint are required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "price": price,
                "tradeMint": trade_mint,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/adrena/close-perp-trade-short",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}
        
        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during close perp trade short: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during close perp trade short: {error}", exc_info=True)
            return {"success": False, "error": str(error)}

    @staticmethod
    def close_perp_trade_long(
        agent: SolanaAgentKit,
        price: float,
        trade_mint: str,
    ) -> Optional[Dict[str, Any]]:
        """
        Closes a perpetual long trade.

        :param agent: An instance of SolanaAgentKit.
        :param price: Trade closing price.
        :param trade_mint: The mint address of the trade asset.
        :return: Transaction signature or error details.
        """
        try:
            if not all([price, trade_mint]):
                raise ValueError("Price and trade_mint are required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "price": price,
                "tradeMint": trade_mint,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/adrena/close-perp-trade-long",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}
        
        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during close perp trade long: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during close perp trade long: {error}", exc_info=True)
            return {"success": False, "error": str(error)}
    
    @staticmethod
    def open_perp_trade_long(
        agent: SolanaAgentKit,
        price: float,
        collateral_amount: float,
        collateral_mint: Optional[str] = None,
        leverage: Optional[float] = None,
        trade_mint: Optional[str] = None,
        slippage: Optional[float] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Opens a perpetual long trade.
        """
        try:
            if not all([price, collateral_amount]):
                raise ValueError("Price and collateral_amount are required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "price": price,
                "collateralAmount": collateral_amount,
                "collateralMint": collateral_mint,
                "leverage": leverage,
                "tradeMint": trade_mint,
                "slippage": slippage,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/adrena/open-perp-trade-long",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}
        
        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during open perp trade long: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during open perp trade long: {error}", exc_info=True)
            return {"success": False, "error": str(error)}

    @staticmethod
    def open_perp_trade_short(
        agent: SolanaAgentKit,
        price: float,
        collateral_amount: float,
        collateral_mint: Optional[str] = None,
        leverage: Optional[float] = None,
        trade_mint: Optional[str] = None,
        slippage: Optional[float] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Opens a perpetual short trade.
        """
        try:
            if not all([price, collateral_amount]):
                raise ValueError("Price and collateral_amount are required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "price": price,
                "collateralAmount": collateral_amount,
                "collateralMint": collateral_mint,
                "leverage": leverage,
                "tradeMint": trade_mint,
                "slippage": slippage,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/adrena/open-perp-trade-short",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}
        
        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during open perp trade short: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during open perp trade short: {error}", exc_info=True)
            return {"success": False, "error": str(error)}


import json
from agentipy.agent import SolanaAgentKit
from langchain.tools import BaseTool

from agentipy.helpers import validate_input

class SolanaTokenDataTool(BaseTool):
    """
    Tool to fetch token data for a given token mint address.
    """
    name:str = "solana_token_data"
    description:str = """Get the token data for a given token mint address.

    Inputs:
    - mintAddress: string, e.g., "So11111111111111111111111111111111111111112" (required)
    """
    solana_kit: SolanaAgentKit

    async def call(self, input: str) -> str:
        try:
            data = json.loads(input)
            schema = {
                "mint_address": {"type": str, "required": True}
            }
            validate_input(data, schema)
            mint_address = data["mint_address"]
            token_data = await self.solana_kit.get_token_data_by_address(mint_address)
            return json.dumps({
                "status": "success",
                "tokenData": token_data,
            })
        except Exception as error:
            return json.dumps({
                "status": "error",
                "message": str(error),
                "code": getattr(error, "code", "UNKNOWN_ERROR"),
            })
        
    def _run(self):
        """Synchronous version of the run method, required by BaseTool."""
        raise NotImplementedError(
            "This tool only supports async execution via _arun. Please use the async interface."
        )

class SolanaTokenDataByTickerTool(BaseTool):
    """
    Tool to fetch token data for a given token ticker.
    """
    name:str = "solana_token_data_by_ticker"
    description:str = """Get the token data for a given token ticker.

    Inputs:
    - ticker: string, e.g., "USDC" (required)
    """
    solana_kit: SolanaAgentKit

    async def call(self, input: str) -> str:
        try:
            data = json.loads(input)
            schema = {
                "ticker": {"type": str, "required": True}
            }
            validate_input(data, schema)

            ticker = data["ticker"]
            token_data = await self.solana_kit.get_token_data_by_ticker(ticker)
            return json.dumps({
                "status": "success",
                "tokenData": token_data,
            })
        except Exception as error:
            return json.dumps({
                "status": "error",
                "message": str(error),
                "code": getattr(error, "code", "UNKNOWN_ERROR"),
            })
        
    def _run(self):
        """Synchronous version of the run method, required by BaseTool."""
        raise NotImplementedError(
            "This tool only supports async execution via _arun. Please use the async interface."
        )



from agentipy.agent import SolanaAgentKit
from agentipy.langchain.land3.land_collection import Create3LandCollectionTool
from agentipy.langchain.land3.land_nft import Create3LandNFTTool




def get_land_tools(solana_kit: SolanaAgentKit):
    return [
        Create3LandCollectionTool(solana_kit=solana_kit),
        Create3LandNFTTool(solana_kit=solana_kit)
    ]





import json
from langchain.tools import BaseTool
from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input


class SolanaSellUsingMoonshotTool(BaseTool):
    name: str = "solana_sell_using_moonshot"
    description:str = """
    Sell a token using Moonshot.

    Input: A JSON string with:
    {
        "mint_str": "string, the mint address of the token to sell",
        "token_balance": 0.01, # optional, token balance to sell (default: 0.01)
        "slippage_bps": 500 # optional, slippage in basis points (default: 500)
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "mint_str": {"type": str, "required": True},
                "token_balance": {"type": float, "required": False, "min": 0},
                "slippage_bps": {"type": int, "required": False, "min": 0, "max": 10000}
            }
            validate_input(data, schema)

            mint_str = data["mint_str"]
            token_balance = data.get("token_balance", 0.01)
            slippage_bps = data.get("slippage_bps", 500)
            
            result = await self.solana_kit.sell_using_moonshot(mint_str, token_balance, slippage_bps)

            return {
                "status": "success",
                "message": "Token sold successfully using Moonshot.",
                "result": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "message": str(e),
                "code": getattr(e, "code", "UNKNOWN_ERROR"),
            }

    def _run(self, input: str):
        """Synchronous version of the run method, required by BaseTool."""
        raise NotImplementedError(
            "This tool only supports async execution via _arun. Please use the async interface."
        )



import json
from langchain.tools import BaseTool
from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input


class LightProtocolSendCompressedAirdropTool(BaseTool):
    name: str = "lightprotocol_send_compressed_airdrop"
    description: str = """
    Sends a compressed airdrop using LightProtocolManager.

    Input: A JSON string with:
    {
        "mint_address": "string, the mint address of the token",
        "amount": "float, the amount to send",
        "decimals": "int, the number of decimal places for the token",
        "recipients": "list, the list of recipient addresses",
        "priority_fee_in_lamports": "int, the priority fee in lamports",
        "should_log": "bool, optional, whether to log the transaction"
    }
    Output:
    {
        "transaction_ids": "list, transaction IDs of the airdrop",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "mint_address": {"type": str, "required": True},
                "amount": {"type": float, "required": True},
                "decimals": {"type": int, "required": True},
                "recipients": {"type": list, "required": True},
                "priority_fee_in_lamports": {"type": int, "required": True},
                "should_log": {"type": bool, "required": False}
            }
            validate_input(data, schema)
            
            transaction_ids = await self.solana_kit.send_compressed_airdrop(
                mint_address=data["mint_address"],
                amount=data["amount"],
                decimals=data["decimals"],
                recipients=data["recipients"],
                priority_fee_in_lamports=data["priority_fee_in_lamports"],
                should_log=data.get("should_log", False)
            )
            return {
                "transaction_ids": transaction_ids,
                "message": "Success"
            }
        except Exception as e:
            return {
                "transaction_ids": None,
                "message": f"Error sending compressed airdrop: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")




import json
import os
from typing import Optional

from anchorpy import Program
from anchorpy import Provider as AnchorProvider
from anchorpy import provider as anchor_provider
from construct import Container
from solana.rpc.async_api import AsyncClient
from solders.pubkey import Pubkey as PublicKey  # type: ignore
from solders.system_program import ID as SYSTEM_PROGRAM_ID
from solders.sysvar import RENT as SYSVAR_RENT_PUBKEY
from spl.token.constants import TOKEN_PROGRAM_ID
from spl.token.instructions import get_associated_token_address

from .constants import LBCLMM_PROGRAM_IDS
from .utils import (bin_id_to_bin_array_index,
                    compute_base_factor_from_fee_bps,
                    derive_bin_array_bitmap_extension,
                    derive_customizable_permissionless_lb_pair, derive_oracle,
                    derive_reserve, is_overflow_default_bin_array_bitmap)


async def create_customizable_permissionless_lb_pair(
    connection: AsyncClient,
    bin_step: int,
    token_x: PublicKey,
    token_y: PublicKey,
    active_id: int,
    fee_bps: int,
    activation_type: str,
    has_alpha_vault: bool,
    creator_key: PublicKey,
    activation_point: Optional[int] = None,
    opt: Optional[dict] = None,
):
    """
    Creates a customizable permissionless LB pair.
    """

    current_dir = os.path.dirname(os.path.abspath(__file__))
    json_path = os.path.join(current_dir, 'idl.json')

    try:
        with open(json_path, 'r') as json_file:
            idl_data = json.load(json_file)
    except FileNotFoundError:
        raise FileNotFoundError(f"IDL file not found at {json_path}")
    except json.JSONDecodeError as e:
        raise ValueError(f"Error decoding JSON from {json_path}: {e}")

    provider = AnchorProvider(connection, {}, anchor_provider.DEFAULT_OPTIONS)
    program_id = opt.get("program_id") if opt else LBCLMM_PROGRAM_IDS[opt.get("cluster")]
    program = Program(idl_data, program_id, provider)

    lb_pair, _ = derive_customizable_permissionless_lb_pair(token_x, token_y, program.program_id)
    reserve_x, _ = derive_reserve(token_x, lb_pair, program.program_id)
    reserve_y, _ = derive_reserve(token_y, lb_pair, program.program_id)
    oracle, _ = derive_oracle(lb_pair, program.program_id)

    active_bin_array_index = bin_id_to_bin_array_index(active_id)
    bin_array_bitmap_extension = (
        derive_bin_array_bitmap_extension(lb_pair, program.program_id)[0]
        if is_overflow_default_bin_array_bitmap(active_bin_array_index)
        else None
    )
    ix_data = Container(
        activeId=active_id,
        binStep=bin_step,
        baseFactor=compute_base_factor_from_fee_bps(bin_step, fee_bps),
        activationType=activation_type,
        activationPoint=activation_point or None,
        hasAlphaVault=has_alpha_vault,
        padding=[0] * 64,
    )

    user_token_x = get_associated_token_address(token_x, creator_key)

    transaction = await program.methods.initialize_customizable_permissionless_lb_pair(ix_data).accounts(
        {
            "lbPair": lb_pair,
            "rent": SYSVAR_RENT_PUBKEY,
            "reserveX": reserve_x,
            "reserveY": reserve_y,
            "binArrayBitmapExtension": bin_array_bitmap_extension,
            "tokenMintX": token_x,
            "tokenMintY": token_y,
            "tokenProgram": TOKEN_PROGRAM_ID,
            "oracle": oracle,
            "systemProgram": SYSTEM_PROGRAM_ID,
            "userTokenX": user_token_x,
            "funder": creator_key,
        }
    ).transaction()

    return transaction



import base64
import json

import aiohttp
from solana.rpc.commitment import Confirmed
from solana.rpc.types import TxOpts
from solders.message import to_bytes_versioned  # type: ignore
from solders.pubkey import Pubkey  # type: ignore
from solders.transaction import VersionedTransaction  # type: ignore

from agentipy.agent import SolanaAgentKit
from agentipy.constants import FLEXLEND_BASE_URL
from agentipy.helpers import fix_asyncio_for_windows

fix_asyncio_for_windows()


class LuloManager:
    BASE_URL = FLEXLEND_BASE_URL

    @staticmethod
    async def lend_asset(agent: SolanaAgentKit, amount: float) -> str:
        """
        Lend tokens for yields using Lulo API.

        Args:
            agent (SolanaAgentKit): SolanaAgentKit instance with connection and wallet.
            amount (float): Amount of USDC to lend.

        Returns:
            str: Transaction signature.
        """
        try:
            url = f"https://blink.lulo.fi/actions?amount={amount}&symbol=USDC"
            headers = {"Content-Type": "application/json"}
            payload = json.dumps({"account": str(agent.wallet.pubkey())})

            async with aiohttp.ClientSession() as session:
                async with session.post(url, headers=headers, data=payload) as response:
                    if response.status != 200:
                        raise Exception(f"Lulo API Error: {response.status}")
                    data = await response.json()

            transaction_bytes = base64.b64decode(data["transaction"])
            lulo_txn = VersionedTransaction.from_bytes(transaction_bytes)

            latest_blockhash = await agent.connection.get_latest_blockhash()

            signature = agent.wallet.sign_message(to_bytes_versioned(lulo_txn.message))

            signed_tx = VersionedTransaction.populate(lulo_txn.message, [signature])

            tx_resp = await agent.connection.send_transaction(
                signed_tx,
                opts=TxOpts(preflight_commitment=Confirmed)
            )
            tx_id = tx_resp.value

            await agent.connection.confirm_transaction(
                tx_id,
                commitment=Confirmed,
                last_valid_block_height=latest_blockhash.value.last_valid_block_height,
            )

            return str(signature)

        except Exception as e:
            raise Exception(f"Lending failed: {str(e)}")

    @staticmethod
    async def lulo_lend(agent: SolanaAgentKit, mint_address: Pubkey, amount: float) -> str:
        """
        Lend tokens for yields using Lulo.

        Args:
            agent (SolanaAgentKit): SolanaAgentKit instance.
            mint_address (Pubkey): SPL Mint address.
            amount (float): Amount to lend.

        Returns:
            str: Transaction signature.
        """
        try:
            url = f"{LuloManager.BASE_URL}/deposit?priorityFee=50000"
            headers = {
                "Content-Type": "application/json",
                "x-wallet-pubkey": str(agent.wallet.pubkey()),
                "x-api-key": agent.FLEXLEND_API_KEY
            }
            payload = json.dumps({
                "owner": str(agent.wallet.pubkey()),
                "mintAddress": str(mint_address),
                "depositAmount": str(amount)
            })

            async with aiohttp.ClientSession() as session:
                async with session.post(url, headers=headers, data=payload) as response:
                    if response.status != 200:
                        raise Exception(f"Lulo API Error: {response.status}")
                    data = await response.json()

            transaction_bytes = base64.b64decode(data["data"]["transactionMeta"][0]["transaction"])
            lulo_txn = VersionedTransaction.from_bytes(transaction_bytes)

            latest_blockhash = await agent.connection.get_latest_blockhash()

            signature = agent.wallet.sign_message(to_bytes_versioned(lulo_txn.message))

            signed_tx = VersionedTransaction.populate(lulo_txn.message, [signature])

            tx_resp = await agent.connection.send_transaction(
                signed_tx,
                opts=TxOpts(preflight_commitment=Confirmed, max_retries=3)
            )
            tx_id = tx_resp.value

            await agent.connection.confirm_transaction(
                tx_id,
                commitment=Confirmed,
                last_valid_block_height=latest_blockhash.value.last_valid_block_height,
            )

            return str(signature)

        except Exception as e:
            raise Exception(f"Lending failed: {str(e)}")

    @staticmethod
    async def lulo_withdraw(agent: SolanaAgentKit, mint_address: Pubkey, amount: float) -> str:
        """
        Withdraw tokens for yields using Lulo.

        Args:
            agent (SolanaAgentKit): SolanaAgentKit instance.
            mint_address (Pubkey): SPL Mint address.
            amount (float): Amount to withdraw.

        Returns:
            str: Transaction signature.
        """
        try:
            if not agent.flexland_api_key:
                raise Exception("Lulo API key not found in agent configuration")

            url = f"{LuloManager.BASE_URL}/withdraw?priorityFee=50000"
            headers = {
                "Content-Type": "application/json",
                "x-wallet-pubkey": str(agent.wallet.pubkey()),
                "x-api-key": agent.flexland_api_key
            }
            payload = json.dumps({
                "owner": str(agent.wallet.pubkey()),
                "mintAddress": str(mint_address),
                "depositAmount": str(amount)
            })

            async with aiohttp.ClientSession() as session:
                async with session.post(url, headers=headers, data=payload) as response:
                    if response.status != 200:
                        raise Exception(f"Lulo API Error: {response.status}")
                    data = await response.json()

            transaction_bytes = base64.b64decode(data["data"]["transactionMeta"][0]["transaction"])
            lulo_txn = VersionedTransaction.from_bytes(transaction_bytes)

            latest_blockhash = await agent.connection.get_latest_blockhash()

            signature = agent.wallet.sign_message(to_bytes_versioned(lulo_txn.message))

            signed_tx = VersionedTransaction.populate(lulo_txn.message, [signature])

            tx_resp = await agent.connection.send_transaction(
                signed_tx,
                opts=TxOpts(preflight_commitment=Confirmed, max_retries=3)
            )
            tx_id = tx_resp.value

            await agent.connection.confirm_transaction(
                tx_id,
                commitment=Confirmed,
                last_valid_block_height=latest_blockhash.value.last_valid_block_height,
            )

            return str(signature)

        except Exception as e:
            raise Exception(f"Withdraw failed: {str(e)}")



class Network:
    def __init__(
            self,
            name: str,
            rpc: list,
            chain_id: int,
            eip1559_support: bool,
            token: str,
            explorer: str,
            decimals: int = 18
    ):
        self.name = name
        self.rpc = rpc
        self.chain_id = chain_id
        self.eip1559_support = eip1559_support
        self.token = token
        self.explorer = explorer
        self.decimals = decimals

    def __repr__(self):
        return f'{self.name}'
    

EthereumRPC = Network(
    name='Ethereum',
    rpc='https://rpc.ankr.com/eth',
    chain_id=1,
    eip1559_support=True,
    token='ETH',
    explorer='https://etherscan.io/'
)

BaseRPC = Network(
    name='Base',
    rpc='https://mainnet.base.org',
    chain_id=8453,
    eip1559_support=False,
    token='ETH',
    explorer='https://basescan.org/'
)

ArbitrumRPC = Network(
    name="Arbitrum",
    rpc='https://rpc.ankr.com/arbitrum/',
    chain_id=42161,
    eip1559_support=True,
    token="ETH",
    explorer='https://arbiscan.io/'
)



import json

from langchain.tools import BaseTool

from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input


class VoltrDepositStrategyTool(BaseTool):
    name: str = "voltr_deposit_strategy"
    description: str = """
    Deposits funds into a Voltr strategy.

    Input: A JSON string with:
    {
        "deposit_amount": "string, the deposit amount",
        "vault": "string, the vault address",
        "strategy": "string, the strategy address"
    }

    Output:
    {
        "transaction_details": "dict, transaction details",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "deposit_amount": {"type": str, "required": True},
                "vault": {"type": str, "required": True},
                "strategy": {"type": str, "required": True},
            }
            validate_input(data, schema)

            deposit_amount = data["deposit_amount"]
            vault = data["vault"]
            strategy = data["strategy"]

            result = await self.solana_kit.deposit_strategy(deposit_amount, vault, strategy)
            return {
                "status": "success",
                "transaction_details": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "message": str(e),
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")


from agentipy.agent import SolanaAgentKit
from agentipy.langchain.raydium.buy import SolanaRaydiumBuyTool
from agentipy.langchain.raydium.sell import SolanaRaydiumSellTool


def get_raydium_tools(solana_kit: SolanaAgentKit):
    return [
        SolanaRaydiumBuyTool(solana_kit=solana_kit),
        SolanaRaydiumSellTool(solana_kit=solana_kit)
    ]




from typing import Dict, List, Optional

from construct import Flag, Int64ul, Struct
from pydantic import BaseModel, field_validator
from solders.pubkey import Pubkey  # type: ignore


class BaseModelWithArbitraryTypes(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = 'ignore'  # Ignore unexpected fields
        validate_assignment = True

class TokenCheck(BaseModelWithArbitraryTypes):
    """
    Model for token report data from RugCheck API.
    """
    mint: Optional[str] = None
    token_program: Optional[str] = None
    token_type: Optional[str] = None
    risks: List[Dict] = []
    score: Optional[int] = None
    creatorTokens: Optional[List] = None

    @field_validator('risks', mode='before')
    @classmethod
    def handle_null_risks(cls, value):
        """Convert null risks to an empty list."""
        return value or []

    def to_user_friendly_string(self) -> str:
        """
        Convert the token report to a user-friendly string.
        """
        return (
            f"Token Report for {self.mint}:\n"
            f"  - Program: {self.token_program}\n"
            f"  - Type: {self.token_type}\n"
            f"  - Risks: {len(self.risks)} risks found\n"
            f"  - Score: {self.score}/100\n"
            f"  - Creator Tokens: {len(self.creatorTokens) if self.creatorTokens else 'None'}"
        )

class Locker(BaseModelWithArbitraryTypes):
    """
    Model for LP locker data.
    """
    programID: str
    tokenAccount: str
    owner: str
    uri: str
    unlockDate: int
    usdcLocked: float
    type: str

    def to_user_friendly_string(self) -> str:
        """
        Convert the locker data to a user-friendly string.
        """
        return (
            f"Locker Details:\n"
            f"  - Program ID: {self.programID}\n"
            f"  - Token Account: {self.tokenAccount}\n"
            f"  - Owner: {self.owner}\n"
            f"  - URI: {self.uri}\n"
            f"  - Unlock Date: {self.unlockDate}\n"
            f"  - USDC Locked: ${self.usdcLocked:,.2f}\n"
            f"  - Type: {self.type}"
        )

class TokenLockers(BaseModelWithArbitraryTypes):
    """
    Model for LP lockers response.
    """
    lockers: Optional[Dict[str, Locker]] = None
    total: Optional[Dict[str, float]] = None

    @field_validator('lockers', 'total', mode='before')
    @classmethod
    def handle_null_fields(cls, value):
        """Convert null fields to empty dictionaries."""
        return value or {}

    def to_user_friendly_string(self) -> str:
        """
        Convert the LP lockers data to a user-friendly string.
        """
        if not self.lockers:
            return "No LP lockers found for this token."
        
        lockers_info = "\n".join(
            f"Locker {i + 1}:\n{locker.to_user_friendly_string()}"
            for i, (_, locker) in enumerate(self.lockers.items())
        )
        total_usdc = self.total.get('totalUSDC', 0) if self.total else 0
        return (
            f"LP Lockers:\n{lockers_info}\n"
            f"Total USDC Locked: ${total_usdc:,.2f}"
        )

class TrendingToken(BaseModelWithArbitraryTypes):
    """
    Model for trending token data.
    """
    mint: str
    vote_count: int
    up_count: int

    def to_user_friendly_string(self) -> str:
        """
        Convert the trending token data to a user-friendly string.
        """
        return (
            f"Trending Token: {self.mint}\n"
            f"  - Votes: {self.vote_count}\n"
            f"  - Up Votes: {self.up_count}"
        )

class Creator(BaseModelWithArbitraryTypes):
    address: str
    percentage: int

class CollectionOptions(BaseModelWithArbitraryTypes):
    name: str
    uri: str
    royalty_basis_points: Optional[int] = None
    creators: Optional[List[Creator]] = None

class CollectionDeployment(BaseModelWithArbitraryTypes):
    collection_address: Pubkey
    signature: bytes

class MintCollectionNFTResponse(BaseModelWithArbitraryTypes):
    mint: Pubkey
    metadata: Pubkey

class PumpfunTokenOptions(BaseModelWithArbitraryTypes):
    twitter: Optional[str] = None
    telegram: Optional[str] = None
    website: Optional[str] = None
    initial_liquidity_sol: Optional[float] = None
    slippage_bps: Optional[int] = None
    priority_fee: Optional[int] = None

class PumpfunLaunchResponse(BaseModelWithArbitraryTypes):
    signature: str
    mint: str
    metadata_uri: Optional[str] = None
    error: Optional[str] = None

class LuloAccountSettings(BaseModelWithArbitraryTypes):
    owner: str
    allowed_protocols: Optional[str] = None
    homebase: Optional[str] = None
    minimum_rate: str

class LuloAccountDetailsResponse(BaseModelWithArbitraryTypes):
    total_value: float
    interest_earned: float
    realtime_apy: float
    settings: LuloAccountSettings

class NetworkPerformanceMetrics(BaseModelWithArbitraryTypes):
    """Data structure for Solana network performance metrics."""
    transactions_per_second: float
    total_transactions: int
    sampling_period_seconds: int
    current_slot: int

class TokenDeploymentResult(BaseModelWithArbitraryTypes):
    """Result of a token deployment operation."""
    mint: Pubkey
    transaction_signature: str

class TokenLaunchResult(BaseModelWithArbitraryTypes):
    """Result of a token launch operation."""
    signature: str
    mint: str
    metadata_uri: str

class TransferResult(BaseModelWithArbitraryTypes):
    """Result of a transfer operation."""
    signature: str
    from_address: str
    to_address: str
    amount: float
    token: Optional[str] = None

class JupiterTokenData(BaseModelWithArbitraryTypes):
    address:str
    symbol:str
    name:str

class GibworkCreateTaskResponse(BaseModelWithArbitraryTypes):
    status: str
    taskId: Optional[str] = None
    signature: Optional[str] = None

class BondingCurveState:
    _STRUCT = Struct(
        "virtual_token_reserves" / Int64ul,
        "virtual_sol_reserves" / Int64ul,
        "real_token_reserves" / Int64ul,
        "real_sol_reserves" / Int64ul,
        "token_total_supply" / Int64ul,
        "complete" / Flag
    )
    def __init__(self, data: bytes) -> None:
        parsed = self._STRUCT.parse(data[8:])


from agentipy.agent import SolanaAgentKit
from agentipy.langchain.meteora.dlmm import SolanaMeteoraDLMMTool


def get_meteora_tools(solana_kit: SolanaAgentKit):
    return [
        SolanaMeteoraDLMMTool(solana_kit=solana_kit)
    ]




from solana.rpc.async_api import AsyncClient
from solders.pubkey import Pubkey  # type: ignore
from spl.token.client import Token
from spl.token.constants import TOKEN_2022_PROGRAM_ID, TOKEN_PROGRAM_ID

from agentipy.agent import SolanaAgentKit


async def get_token_decimals(agent: SolanaAgentKit, mint: Pubkey) -> int:
    """
    Get the number of decimals for a given token mint address.

    Args:
        agent (SolanaAgentKit): The Solana agent instance.
        mint (Pubkey): The mint address of the token.

    Returns:
        int: The number of decimals for the token.

    Raises:
        Exception: If fetching mint info fails for both standard and Token 2022 program.
    """
    try:
        token_client = Token(agent.connection, mint, TOKEN_PROGRAM_ID, agent.wallet)
        mint_info = await token_client.get_mint_info()
        return mint_info.decimals
    except Exception:
        try:
            token_client = Token(agent.connection, mint, TOKEN_2022_PROGRAM_ID, agent.wallet.payer)
            mint_info = await token_client.get_mint_info()
            return mint_info.decimals
        except Exception as final_error:
            raise Exception(f"Failed to fetch mint info for token {mint}: {str(final_error)}")



from langchain.tools import BaseTool
from agentipy.agent import SolanaAgentKit

class RugCheckFetchRecentlyVerifiedTokensTool(BaseTool):
    name: str = "rugcheck_fetch_recently_verified_tokens"
    description: str = """
    Fetches recently verified tokens using RugCheckManager.

    Input: None
    Output:
    {
        "recently_verified_tokens": "list, a list of recently verified tokens",
        "message": "string, if an error occurs"
    }
    """
    agent_kit: SolanaAgentKit

    async def _arun(self):
        try:
            recently_verified_tokens = await self.agent_kit.fetch_recently_verified_tokens()
            return {
                "recently_verified_tokens": recently_verified_tokens,
                "message": "Success"
            }
        except Exception as e:
            return {
                "recently_verified_tokens": None,
                "message": f"Error fetching recently verified tokens: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")



import base64
import json

import aiohttp
from solana.rpc.commitment import Confirmed
from solana.rpc.types import TxOpts
from solders.message import to_bytes_versioned  # type: ignore
from solders.pubkey import Pubkey  # type: ignore
from solders.transaction import VersionedTransaction  # type: ignore

from agentipy.agent import SolanaAgentKit


class SolayerManager:
    BASE_URL = "https://app.solayer.org/api/action/restake/ssol"

    @staticmethod
    async def stake_with_solayer(agent: SolanaAgentKit, amount: float) -> str:
        """
        Stake SOL with Solayer.

        Args:
            agent (SolanaAgentKit): SolanaAgentKit instance.
            amount (float): Amount of SOL to stake.

        Returns:
            str: Transaction signature.
        """
        try:
            url = f"{SolayerManager.BASE_URL}?amount={amount}"
            headers = {"Content-Type": "application/json"}
            payload = json.dumps({"account": str(agent.wallet_address)})

            async with aiohttp.ClientSession() as session:
                async with session.post(url, headers=headers, data=payload) as response:
                    if response.status != 200:
                        error_data = await response.json()
                        raise Exception(error_data.get("message", "Staking request failed"))
                    data = await response.json()

            transaction_bytes = base64.b64decode(data["transaction"])
            txn = VersionedTransaction.from_bytes(transaction_bytes)

            latest_blockhash = await agent.connection.get_latest_blockhash()

            signature = agent.wallet.sign_message(to_bytes_versioned(txn.message))
            signed_tx = VersionedTransaction.populate(txn.message, [signature])

            tx_resp = await agent.connection.send_transaction(
                signed_tx,
                opts=TxOpts(preflight_commitment=Confirmed, max_retries=3)
            )
            tx_id = tx_resp.value

            await agent.connection.confirm_transaction(
                tx_id,
                commitment=Confirmed,
                last_valid_block_height=latest_blockhash.value.last_valid_block_height,
            )

            return str(signature)

        except Exception as e:
            raise Exception(f"Solayer sSOL staking failed: {str(e)}")



import json
from agentipy.agent import SolanaAgentKit
from langchain.tools import BaseTool
from agentipy.helpers import validate_input

class BackpackGetMarketsTool(BaseTool):
    name: str = "backpack_get_markets"
    description: str = """
    Fetches all markets using the BackpackManager.

    Input: None
    Output:
    {
        "markets": "list, the available markets",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self):
        try:
            markets = await self.solana_kit.get_markets()
            return {
                "markets": markets,
                "message": "Success"
            }
        except Exception as e:
            return {
                "markets": None,
                "message": f"Error fetching markets: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetMarketTool(BaseTool):
    name: str = "backpack_get_market"
    description: str = """
    Fetches a specific market using the BackpackManager.

    Input: A JSON string with market query parameters.
    Output:
    {
        "market": "dict, the market data",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            market = await self.solana_kit.get_market(**data)
            return {
                "market": market,
                "message": "Success"
            }
        except Exception as e:
            return {
                "market": None,
                "message": f"Error fetching market: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetTickersTool(BaseTool):
    name: str = "backpack_get_tickers"
    description: str = """
    Fetches tickers for all markets using the BackpackManager.

    Input: None
    Output:
    {
        "tickers": "list, the market tickers",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self):
        try:
            tickers = await self.solana_kit.get_tickers()
            return {
                "tickers": tickers,
                "message": "Success"
            }
        except Exception as e:
            return {
                "tickers": None,
                "message": f"Error fetching tickers: {str(e)}"
            }

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetDepthTool(BaseTool):
    name: str = "backpack_get_depth"
    description: str = """
    Fetches the order book depth for a given market symbol using the BackpackManager.

    Input: A JSON string with:
    {
        "symbol": "string, the market symbol"
    }
    Output:
    {
        "depth": "dict, the order book depth",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "symbol": {"type": str, "required": True}
            }
            validate_input(data, schema)

            symbol = data["symbol"]
            depth = await self.solana_kit.get_depth(symbol)
            return {
                "depth": depth,
                "message": "Success"
            }
        except Exception as e:
            return {
                "depth": None,
                "message": f"Error fetching depth: {str(e)}"
            }

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetKlinesTool(BaseTool):
    name: str = "backpack_get_klines"
    description: str = """
    Fetches K-Lines data for a given market symbol using the BackpackManager.

    Input: A JSON string with:
    {
        "symbol": "string, the market symbol",
        "interval": "string, the interval for K-Lines",
        "start_time": "int, the start time for data",
        "end_time": "int, optional, the end time for data"
    }
    Output:
    {
        "klines": "dict, the K-Lines data",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            schema = {
                "symbol": {"type": str, "required": True},
                "interval": {"type": str, "required": True},
                "start_time": {"type": int, "required": True}
            }
            data = json.loads(input)
            validate_input(data, schema)

            klines = await self.solana_kit.get_klines(
                symbol=data["symbol"],
                interval=data["interval"],
                start_time=data["start_time"],
                end_time=data.get("end_time")
            )
            return {
                "klines": klines,
                "message": "Success"
            }
        except Exception as e:
            return {
                "klines": None,
                "message": f"Error fetching K-Lines: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetMarkPriceTool(BaseTool):
    name: str = "backpack_get_mark_price"
    description: str = """
    Fetches mark price, index price, and funding rate for a given market symbol.

    Input: A JSON string with:
    {
        "symbol": "string, the market symbol"
    }
    Output:
    {
        "mark_price_data": "dict, the mark price data",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            schema = {
                "symbol": {"type": str, "required": True}
            }
            data = json.loads(input)
            validate_input(data, schema)

            symbol = data["symbol"]
            mark_price_data = await self.solana_kit.get_mark_price(symbol)
            return {
                "mark_price_data": mark_price_data,
                "message": "Success"
            }
        except Exception as e:
            return {
                "mark_price_data": None,
                "message": f"Error fetching mark price: {str(e)}"
            }

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetOpenInterestTool(BaseTool):
    name: str = "backpack_get_open_interest"
    description: str = """
    Fetches the open interest for a given market symbol using the BackpackManager.

    Input: A JSON string with:
    {
        "symbol": "string, the market symbol"
    }
    Output:
    {
        "open_interest": "dict, the open interest data",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            schema = {
                "symbol": {"type": str, "required": True}
            }
            data = json.loads(input)
            validate_input(data, schema)

            symbol = data["symbol"]
            open_interest = await self.solana_kit.get_open_interest(symbol)
            return {
                "open_interest": open_interest,
                "message": "Success"
            }
        except Exception as e:
            return {
                "open_interest": None,
                "message": f"Error fetching open interest: {str(e)}"
            }

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetFundingIntervalRatesTool(BaseTool):
    name: str = "backpack_get_funding_interval_rates"
    description: str = """
    Fetches funding interval rate history for futures using the BackpackManager.

    Input: A JSON string with:
    {
        "symbol": "string, the market symbol",
        "limit": "int, optional, maximum results to return (default: 100)",
        "offset": "int, optional, records to skip (default: 0)"
    }
    Output:
    {
        "funding_rates": "dict, the funding interval rate data",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            schema = {
                "symbol": {"type": str, "required": True},
                "limit": {"type": int, "required": False},
                "offset": {"type": int, "required": False}
            }
            data = json.loads(input)
            validate_input(data, schema)

            symbol = data["symbol"]
            limit = data.get("limit", 100)
            offset = data.get("offset", 0)

            funding_rates = await self.solana_kit.get_funding_interval_rates(
                symbol=symbol,
                limit=limit,
                offset=offset
            )
            return {
                "funding_rates": funding_rates,
                "message": "Success"
            }
        except Exception as e:
            return {
                "funding_rates": None,
                "message": f"Error fetching funding interval rates: {str(e)}"
            }

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")



import json
from langchain.tools import BaseTool
from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input

class RugCheckFetchAllDomainsTool(BaseTool):
    name: str = "rugcheck_fetch_all_domains"
    description: str = """
    Fetches all registered domains with optional pagination and filtering using RugCheckManager.

    Input: A JSON string with:
    {
        "page": "int, optional, the page number for pagination (default: 1)",
        "limit": "int, optional, the number of records per page (default: 50)",
        "verified": "bool, optional, filter for verified domains (default: False)"
    }
    Output:
    {
        "domains": "list, a list of all registered domains",
        "message": "string, if an error occurs"
    }
    """
    agent_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "page": {"type": int, "required": False},
                "limit": {"type": int, "required": False},
                "verified": {"type": bool, "required": False}
            }
            validate_input(data, schema)
            domains = await self.agent_kit.fetch_all_domains(
                page=data.get("page", 1),
                limit=data.get("limit", 50),
                verified=data.get("verified", False)
            )
            return {
                "domains": domains,
                "message": "Success"
            }
        except Exception as e:
            return {
                "domains": None,
                "message": f"Error fetching all domains: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")



from enum import Enum
import logging
import os
from typing import Optional, Union

from eth_account import Account
from web3 import Web3

from agentipy.constants import API_VERSION, BASE_PROXY_URL
from agentipy.tools.evm.wallet_opts import Web3EVMClient
from agentipy.utils import AgentKitError
from agentipy.utils.evm.general.networks import Network
from allora_sdk.v2.api_client import (
    PriceInferenceTimeframe,
    PriceInferenceToken,
    SignatureFormat,
)

from agentipy.wallet.crossMint_wallet_client import CrossmintWalletClient
from agentipy.wallet.evm_wallet_client import EVMWalletClient

logger = logging.getLogger(__name__)


class WalletType(str, Enum):
    """Enum for wallet types"""

    PRIVATE_KEY = "private_key"
    CROSSMINT = "crossmint"


class EvmAgentKit:
    """
    Main class for interacting with multiple EVM blockchains.
    Supports token operations, contract interactions, and chain-specific functionality.

    Attributes:
        web3 (Web3): Web3 provider for interacting with the blockchain.
        wallet_address (str): Public address of the wallet.
        private_key (str): Private key for signing transactions.
        chain_id (int): Chain ID of the connected EVM network.
        token (str): Native token symbol of the network.
        explorer (str): Blockchain explorer URL.
    """

    def __init__(
        self,
        network: Network,
        wallet_type: Union[WalletType, str] = WalletType.PRIVATE_KEY,
        private_key: Optional[str] = None,
        rpc_url: Optional[str] = None,
        rpc_api_key: Optional[str] = None,
        crossmint_api_key: Optional[str] = None,
        crossmint_wallet_locator: Optional[str] = None,
        openai_api_key: Optional[str] = None,
        coingecko_api_key: Optional[str] = None,
        coingecko_demo_api_key: Optional[str] = None,
        stork_api_key: Optional[str] = None,
        elfa_ai_api_key: Optional[str] = None,
        allora_api_key: Optional[str] = None,
        generate_wallet: bool = False,
    ):
        self.network = network
        self.wallet_type = wallet_type
        self.rpc_url = rpc_url or os.getenv("EVM_RPC_URL", network.rpc)
        self.rpc_api_key = rpc_api_key or os.getenv("EVM_RPC_API_KEY", "")
        self.web3 = Web3(Web3.HTTPProvider(self.rpc_url))
        self.openai_api_key = openai_api_key or os.getenv("OPENAI_API_KEY", "")
        self.coingecko_api_key = coingecko_api_key or os.getenv(
            "COINGECKO_PRO_API_KEY", ""
        )
        self.coingecko_demo_api_key = coingecko_demo_api_key or os.getenv(
            "COINGECKO_DEMO_API_KEY", ""
        )
        self.stork_api_key = stork_api_key or os.getenv("STORK_API_KEY", "")
        self.elfa_ai_api_key = elfa_ai_api_key or os.getenv("ELFA_AI_API_KEY", "")
        self.allora_api_key = allora_api_key or os.getenv("ALLORA_API_KEY", "")
        self.chain_id = network.chain_id
        self.token = network.token
        self.explorer = network.explorer
        self.eip1559_support = network.eip1559_support
        self.base_proxy_url = BASE_PROXY_URL
        self.api_version = API_VERSION

        if wallet_type == WalletType.CROSSMINT or wallet_type == "crossmint":
            # Use Crossmint wallet
            self.crossmint_api_key = crossmint_api_key or os.getenv(
                "CROSSMINT_API_KEY", ""
            )
            if not self.crossmint_api_key:
                raise AgentKitError(
                    "A valid Crossmint API key must be provided when using Crossmint wallet."
                )

            # Initialize the Crossmint wallet client with locator
            self.wallet_client = CrossmintWalletClient(
                self.web3,
                self.crossmint_api_key,
                wallet_locator=crossmint_wallet_locator,
            )
            self.wallet_address = self.wallet_client.get_address()
            self.private_key = None  # No private key when using Crossmint

        else:  # Default to private key wallet
            # Handle private key wallet initialization
            if generate_wallet:
                self.private_key, self.wallet_address = self.generate_wallet()
            else:
                self.private_key = private_key or os.getenv("EVM_PRIVATE_KEY", "")
                if not self.private_key:
                    raise AgentKitError(
                        "A valid private key must be provided when using private key wallet."
                    )
                self.wallet_address = self.web3.eth.account.from_key(
                    self.private_key
                ).address

            self.wallet_client = EVMWalletClient(self.web3, self.private_key)

        logger.info(
            f"Connected to {network.name} (Chain ID: {self.chain_id}) - RPC: {self.rpc_url}"
        )
        logger.info(f"Wallet address: {self.wallet_address}")

    @staticmethod
    def generate_wallet():
        """
        Generates a new EVM wallet with a random private key.
        """
        account = Account.create()
        private_key = account.key.hex()
        wallet_address = account.address
        logger.info(f"New Wallet Generated: {wallet_address}")
        return private_key, wallet_address

    async def get_sentient_listings(
        self, page_number: Optional[int] = 1, page_size: Optional[int] = 30
    ):
        """
        Retrieves Sentient listings.

        Args:
            page_number (int, optional): The page number for paginated results (default: 1).
            page_size (int, optional): The number of items per page (default: 30).

        Returns:
            dict: Listings data or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.get_sentient_listings(self, page_number, page_size)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch Sentient listings: {e}")

    async def buy_sentient(
        self, token_address: str, amount: str, builder_id: Optional[int] = None
    ):
        """
        Buys Sentient tokens.

        Args:
            token_address (str): The token address.
            amount (str): The amount to purchase.
            builder_id (int, optional): The builder ID for the purchase.

        Returns:
            dict: Transaction receipt or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.buy_sentient(self, token_address, amount, builder_id)
        except Exception as e:
            raise AgentKitError(f"Failed to buy Sentient tokens: {e}")

    async def sell_sentient(
        self, token_address: str, amount: str, builder_id: Optional[int] = None
    ):
        """
        Sells Sentient tokens.

        Args:
            token_address (str): The token address.
            amount (str): The amount to sell.
            builder_id (int, optional): The builder ID for the sale.

        Returns:
            dict: Transaction receipt or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.sell_sentient(
                self, token_address, amount, builder_id
            )
        except Exception as e:
            raise AgentKitError(f"Failed to sell Sentient tokens: {e}")

    async def buy_prototype(
        self,
        token_address: str,
        amount: str,
        builder_id: Optional[int] = None,
        slippage: Optional[float] = None,
    ):
        """
        Buys Prototype tokens.

        Args:
            token_address (str): The token address.
            amount (str): The amount to purchase.
            builder_id (int, optional): The builder ID for the purchase.
            slippage (float, optional): Slippage tolerance percentage.

        Returns:
            dict: Transaction receipt or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.buy_prototype(
                self, token_address, amount, builder_id, slippage
            )
        except Exception as e:
            raise AgentKitError(f"Failed to buy Prototype tokens: {e}")

    async def sell_prototype(
        self,
        token_address: str,
        amount: str,
        builder_id: Optional[int] = None,
        slippage: Optional[float] = None,
    ):
        """
        Sells Prototype tokens.

        Args:
            token_address (str): The token address.
            amount (str): The amount to sell.
            builder_id (int, optional): The builder ID for the sale.
            slippage (float, optional): Slippage tolerance percentage.

        Returns:
            dict: Transaction receipt or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.sell_prototype(
                self, token_address, amount, builder_id, slippage
            )
        except Exception as e:
            raise AgentKitError(f"Failed to sell Prototype tokens: {e}")

    async def check_sentient_allowance(
        self, amount: str, from_token_address: Optional[str] = None
    ):
        """
        Checks Sentient token allowance.

        Args:
            amount (str): The amount to check allowance for.
            from_token_address (str, optional): The address of the token being checked.

        Returns:
            dict: Boolean indicating whether allowance is sufficient.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.check_sentient_allowance(
                self, amount, from_token_address
            )
        except Exception as e:
            raise AgentKitError(f"Failed to check Sentient allowance: {e}")

    async def approve_sentient_allowance(
        self, amount: str, from_token_address: Optional[str] = None
    ):
        """
        Approves Sentient token allowance.

        Args:
            amount (str): The amount to approve.
            from_token_address (str, optional): The token address being approved.

        Returns:
            dict: Transaction hash or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.approve_sentient_allowance(
                self, amount, from_token_address
            )
        except Exception as e:
            raise AgentKitError(f"Failed to approve Sentient allowance: {e}")

    async def check_prototype_allowance(
        self, amount: str, from_token_address: Optional[str] = None
    ):
        """
        Checks Prototype token allowance.

        Args:
            amount (str): The amount to check allowance for.
            from_token_address (str, optional): The address of the token being checked.

        Returns:
            dict: Boolean indicating whether allowance is sufficient.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.check_prototype_allowance(
                self, amount, from_token_address
            )
        except Exception as e:
            raise AgentKitError(f"Failed to check Prototype allowance: {e}")

    async def approve_prototype_allowance(
        self, amount: str, from_token_address: Optional[str] = None
    ):
        """
        Approves Prototype token allowance.

        Args:
            amount (str): The amount to approve.
            from_token_address (str, optional): The token address being approved.

        Returns:
            dict: Transaction hash or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.approve_prototype_allowance(
                self, amount, from_token_address
            )
        except Exception as e:
            raise AgentKitError(f"Failed to approve Prototype allowance: {e}")

    async def get_prototype_listing(self, page_number: int = 1, page_size: int = 30):
        """
        Retrieves Prototype token listings.

        Args:
            page_number (int, optional): Page number for pagination.
            page_size (int, optional): Number of items per page.

        Returns:
            dict: List of Prototype token listings or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.get_prototype_listing(self, page_number, page_size)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch Prototype listings: {e}")

    async def fetch_klines(
        self, token_address: str, granularity: int, start: int, end: int, limit: int
    ):
        """
        Fetches Klines (candlestick chart data) for a token.

        Args:
            token_address (str): The token address.
            granularity (int): The granularity of the data.
            start (int): The start timestamp.
            end (int): The end timestamp.
            limit (int): The number of data points.

        Returns:
            dict: Kline data or error details.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.fetch_klines(
                self, token_address, granularity, start, end, limit
            )
        except Exception as e:
            raise AgentKitError(f"Failed to fetch Klines: {e}")

    async def search_virtual_token_by_keyword(self, keyword: str):
        """
        Searches for a virtual token by keyword.

        Args:
            keyword (str): The search keyword.

        Returns:
            dict: Token details or error message.
        """
        from agentipy.tools.evm.use_virtuals import VirtualsManager

        try:
            return VirtualsManager.search_virtual_token_by_keyword(self, keyword)
        except Exception as e:
            raise AgentKitError(f"Failed to search virtual token by keyword: {e}")

    async def get_uniswap_quote(
        self,
        input_token_address: str,
        output_token_address: str,
        amount_in_raw: str,
        input_token_decimals: int = 18,
        output_token_decimals: int = 18,
        slippage: float = 0.5,
        fee_amount: Optional[int] = 3000,
    ):
        """
        Retrieves a quote from Uniswap for swapping between two tokens.

        Args:
            input_token_address (str): The address of the input token.
            output_token_address (str): The address of the output token.
            amount_in_raw (str): The raw amount of input token to swap.
            input_token_decimals (int, optional): The number of decimals for the input token. Defaults to 18.
            output_token_decimals (int, optional): The number of decimals for the output token. Defaults to 18.
            slippage (float, optional): The slippage tolerance percentage. Defaults to 0.5.
            fee_amount (int, optional): The fee amount for the pool. Defaults to 3000 (MEDIUM).

        Returns:
            dict: Quote data or error details.
        """
        if self.network.name != "Base":
            raise AgentKitError("This function is only available for Base network.")

        from agentipy.tools.evm.use_uniswap import UniswapManager

        try:
            return UniswapManager.get_quote(
                self,
                input_token_address,
                output_token_address,
                amount_in_raw,
                input_token_decimals,
                output_token_decimals,
                slippage,
                fee_amount,
            )
        except Exception as e:
            raise AgentKitError(f"Failed to get Uniswap quote: {e}")

    async def trade_on_uniswap(
        self,
        input_token_address: str,
        output_token_address: str,
        amount_in_raw: str,
        input_token_decimals: int = 18,
        output_token_decimals: int = 18,
        slippage: float = 0.5,
        fee_amount: Optional[int] = 3000,
    ):
        """
        Executes a token swap on Uniswap.

        Args:
            input_token_address (str): The address of the input token.
            output_token_address (str): The address of the output token.
            amount_in_raw (str): The raw amount of input token to swap.
            input_token_decimals (int, optional): The number of decimals for the input token. Defaults to 18.
            output_token_decimals (int, optional): The number of decimals for the output token. Defaults to 18.
            slippage (float, optional): The slippage tolerance percentage. Defaults to 0.5.
            fee_amount (int, optional): The fee amount for the pool. Defaults to 3000 (MEDIUM).

        Returns:
            dict: Transaction details or error information.
        """

        if self.network.name != "Base":
            raise AgentKitError("This function is only available for Base network.")

        from agentipy.tools.evm.use_uniswap import UniswapManager

        try:
            return UniswapManager.trade(
                self,
                input_token_address,
                output_token_address,
                amount_in_raw,
                input_token_decimals,
                output_token_decimals,
                slippage,
                fee_amount,
            )
        except Exception as e:
            raise AgentKitError(f"Failed to execute Uniswap trade: {e}")

    async def get_trending_tokens(self):
        """
        Get trending tokens from CoinGecko.

        Returns:
            dict: Trending tokens data.
        """
        from agentipy.tools.use_coingecko import CoingeckoManager

        try:
            return await CoingeckoManager.get_trending_tokens(self)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch trending tokens: {e}")

    async def get_coin_price_vs(
        self, coin_ids: list[str], vs_currencies: list[str] = ["usd"]
    ):
        """
        Get token price data from CoinGecko.

        Args:
            coin_ids (list[str]): A list of token contract addresses.
            vs_currencies (list[str], optional): A list of currency codes for price comparison. Default is ["usd"].

        Returns:
            dict: Token price data from CoinGecko.
        """
        from agentipy.tools.use_coingecko import CoingeckoManager

        try:
            return await CoingeckoManager.get_coin_price_vs(
                self, coin_ids, vs_currencies
            )
        except Exception as e:
            raise AgentKitError(f"Failed to fetch token price data: {e}")

    async def get_trending_pools(self, duration: str = "24h"):
        """
        Get trending pools from CoinGecko for the Solana network.

        Args:
            duration (str): Duration filter for trending pools. Allowed values: "5m", "1h", "6h", "24h". Default is "24h".

        Returns:
            dict: Trending pools data.
        """
        from agentipy.tools.use_coingecko import CoingeckoManager

        try:
            return await CoingeckoManager.get_trending_pools(self, duration)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch trending pools: {e}")

    async def get_top_gainers(
        self, duration: str = "24h", top_coins: int | str = "all"
    ):
        """
        Get top gainers from CoinGecko.

        Args:
            duration (str): Duration filter for top gainers. Default is "24h".
            top_coins (int or str): The number of top coins to return. Default is "all".

        Returns:
            dict: Top gainers data.
        """
        from agentipy.tools.use_coingecko import CoingeckoManager

        try:
            return await CoingeckoManager.get_top_gainers(self, duration, top_coins)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch top gainers: {e}")

    async def get_token_price_data(self, token_addresses: list[str]):
        """
        Get token price data for a list of token addresses from CoinGecko.

        Args:
            token_addresses (list[str]): A list of token contract addresses.

        Returns:
            dict: Token price data from CoinGecko.
        """
        from agentipy.tools.use_coingecko import CoingeckoManager

        try:
            return await CoingeckoManager.get_token_price_data(self, token_addresses)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch token price data: {e}")

    async def get_token_info(self, token_address: str):
        """
        Get token info for a given token address from CoinGecko.

        Args:
            token_address (str): The token's contract address.

        Returns:
            dict: Token info data.
        """
        from agentipy.tools.use_coingecko import CoingeckoManager

        try:
            return await CoingeckoManager.get_token_info(self, token_address)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch token info: {e}")

    async def get_latest_pools(self):
        """
        Get the latest pools from CoinGecko for the Solana network.

        Returns:
            dict: Latest pools data.
        """
        from agentipy.tools.use_coingecko import CoingeckoManager

        try:
            return await CoingeckoManager.get_latest_pools(self)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch latest pools: {e}")

    async def stork_fetch_price(self, asset_id: str):
        from agentipy.tools.use_stork import StorkManager

        try:
            return await StorkManager.get_price(self, asset_id)
        except Exception as e:
            raise AgentKitError(f"Failed to {e}")

    async def ping_elfa_ai_api(self) -> dict:
        """
        Ping the Elfa AI API.

        Returns:
            dict: API response.
        """
        from agentipy.tools.use_elfa_ai import ElfaAiManager

        try:
            return await ElfaAiManager.ping_elfa_ai_api(self)
        except Exception as e:
            raise AgentKitError(f"Failed to ping Elfa AI API: {e}")

    async def get_elfa_ai_api_key_status(self) -> dict:
        """
        Get the Elfa AI API key status.

        Returns:
            dict: API key status.
        """
        from agentipy.tools.use_elfa_ai import ElfaAiManager

        try:
            return await ElfaAiManager.get_elfa_ai_api_key_status(self)
        except Exception as e:
            raise AgentKitError(f"Failed to get Elfa AI API key status: {e}")

    async def get_smart_mentions(self, limit: int = 100, offset: int = 0) -> dict:
        """
        Get smart mentions from Elfa AI.

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            limit (int): Number of mentions to retrieve.
            offset (int): Offset for pagination.

        Returns:
            dict: Mentions data.
        """
        from agentipy.tools.use_elfa_ai import ElfaAiManager

        try:
            return await ElfaAiManager.get_smart_mentions(self, limit, offset)
        except Exception as e:
            raise AgentKitError(f"Failed to get smart mentions: {e}")

    async def get_top_mentions_by_ticker(
        self,
        ticker: str,
        time_window: str = "1h",
        page: int = 1,
        page_size: int = 10,
        include_account_details: bool = False,
    ) -> dict:
        """
        Get top mentions by ticker.

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            ticker (str): The ticker symbol.
            time_window (str): The time window for mentions.
            page (int): Page number.
            page_size (int): Number of results per page.
            include_account_details (bool): Whether to include account details.

        Returns:
            dict: Mentions data.
        """
        from agentipy.tools.use_elfa_ai import ElfaAiManager

        try:
            return await ElfaAiManager.get_top_mentions_by_ticker(
                self, ticker, time_window, page, page_size, include_account_details
            )
        except Exception as e:
            raise AgentKitError(f"Failed to get top mentions by ticker: {e}")

    async def search_mentions_by_keywords(
        self,
        keywords: str,
        from_timestamp: int,
        to_timestamp: int,
        limit: int = 20,
        cursor: str = None,
    ) -> dict:
        """
        Search mentions by keywords.

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            keywords (str): Keywords for search.
            from_timestamp (int): Start timestamp.
            to_timestamp (int): End timestamp.
            limit (int): Number of results to fetch.
            cursor (str): Optional cursor for pagination.

        Returns:
            dict: Search results.
        """
        from agentipy.tools.use_elfa_ai import ElfaAiManager

        try:
            return await ElfaAiManager.search_mentions_by_keywords(
                self, keywords, from_timestamp, to_timestamp, limit, cursor
            )
        except Exception as e:
            raise AgentKitError(f"Failed to search mentions by keywords: {e}")

    async def get_trending_tokens_using_elfa_ai(
        self,
        time_window: str = "24h",
        page: int = 1,
        page_size: int = 50,
        min_mentions: int = 5,
    ) -> dict:
        """
        Get trending tokens using Elfa AI.

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            time_window (str): Time window for trending tokens.
            page (int): Page number.
            page_size (int): Number of results per page.
            min_mentions (int): Minimum number of mentions required.

        Returns:
            dict: Trending tokens data.
        """
        from agentipy.tools.use_elfa_ai import ElfaAiManager

        try:
            return await ElfaAiManager.get_trending_tokens_using_elfa_ai(
                self, time_window, page, page_size, min_mentions
            )
        except Exception as e:
            raise AgentKitError(f"Failed to get trending tokens using Elfa AI: {e}")

    async def get_smart_twitter_account_stats(self, username: str) -> dict:
        """
        Get smart Twitter account stats.

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            username (str): The Twitter username.

        Returns:
            dict: Account statistics data.
        """
        from agentipy.tools.use_elfa_ai import ElfaAiManager

        try:
            return await ElfaAiManager.get_smart_twitter_account_stats(self, username)
        except Exception as e:
            raise AgentKitError(f"Failed to get smart Twitter account stats: {e}")

    async def get_price_prediction(
        self,
        asset: PriceInferenceToken,
        timeframe: PriceInferenceTimeframe,
        signature_format: SignatureFormat = SignatureFormat.ETHEREUM_SEPOLIA,
    ):
        """
        Fetch a future price prediction for BTC or ETH for a given timeframe (5m or 8h) from the Allora Network.

        :param ticker: The crypto asset symbol (e.g., "BTC" or "ETH").
        :param timeframe: The prediction timeframe ("5m" or "8h").
        :return: A dictionary containing the predicted price and confidence interval.
        """
        from agentipy.tools.use_allora import AlloraManager

        try:
            return await AlloraManager.get_price_prediction(
                self, asset, timeframe, signature_format
            )
        except Exception as e:
            raise AgentKitError(f"Failed to fetch price prediction: {e}")

    async def get_inference_by_topic_id(self, topic_id: int):
        """
        Fetch a price inference for BTC or ETH for a given timeframe (5m or 8h) from the Allora Network.

        :param ticker: The crypto asset symbol (e.g., "BTC" or "ETH").
        :param timeframe: The prediction timeframe ("5m" or "8h").
        :return: A dictionary containing the predicted price and confidence interval.
        """
        from agentipy.tools.use_allora import AlloraManager

        try:
            return await AlloraManager.get_inference_by_topic_id(self, topic_id)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch price inference: {e}")

    async def get_all_topics(self):
        """
        Fetch all topics from the Allora Network.

        :return: A list of topic IDs.
        """
        from agentipy.tools.use_allora import AlloraManager

        try:
            return await AlloraManager.get_all_topics(self)
        except Exception as e:
            raise AgentKitError(f"Failed to fetch all topics: {e}")

    async def pyth_fetch_price(self, base_token_ticker: str, quote_token_ticker: str):
        from agentipy.tools.use_pyth import PythManager

        try:
            return await PythManager.get_price(
                self, base_token_ticker, quote_token_ticker
            )
        except Exception as e:
            raise AgentKitError(f"Failed to {e}")



import json
from langchain.tools import BaseTool
from agentipy.agent import SolanaAgentKit



class GetAllDomainsTLDsTool(BaseTool):
    name: str = "get_all_domains_tlds"
    description: str = """
    Retrieves all available top-level domains (TLDs).

    Input: No input required.
    Output:
    {
        "tlds": "list of strings, available TLDs",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self):
        try:
            tlds = await self.solana_kit.get_all_domains_tlds()
            return {"tlds": tlds, "message": "Success"} if tlds else {"message": "No TLDs found"}
        except Exception as e:
            return {"message": f"Error fetching TLDs: {str(e)}"}

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")



import json
from langchain.tools import BaseTool
from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input


class SolanaFaucetTool(BaseTool):
    name:str = "solana_request_funds"
    description:str = "Request test funds from a Solana faucet."
    solana_kit: SolanaAgentKit

    async def _arun(self):
        try:
            result = await self.solana_kit.request_faucet_funds()
            return {
                "status": "success",
                "message": "Faucet funds requested successfully",
                "result": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "message": str(e),
                "code": getattr(e, "code", "UNKNOWN_ERROR"),
            }

    def _run(self):
        """Synchronous version of the run method, required by BaseTool."""
        raise NotImplementedError(
            "This tool only supports async execution via _arun. Please use the async interface."
        )



from agentipy.utils.evm.general.networks import BaseRPC


class Token:
    def __init__(
            self,
            chainId: int,
            address: list,
            decimals: int = 18
    ):
        self.chainId = chainId
        self.address = address
        self.decimals = decimals

    def __repr__(self):
        return f'{self.name}'

WETH_BASE = Token(
    chainId=BaseRPC,
    address="0x4200000000000000000000000000000000000006",
    decimals=18
)

USDC_BASE = Token(
    chainId=BaseRPC,
    address="0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
    decimals=6
)


from langchain.tools import BaseTool
from agentipy.agent import SolanaAgentKit
class AlloraGetAllTopicsTool(BaseTool):
    name: str = "allora_get_all_topics"
    description: str = """
    Fetches all topics from the Allora Network using AlloraManager.

    Input: None
    Output:
    {
        "topics": "list, the list of topic IDs",
        "message": "string, if an error occurs"
    }
    """
    agent_kit: SolanaAgentKit

    async def _arun(self):
        try:
            topics = await self.agent_kit.get_all_topics()
            return {
                "topics": topics,
                "message": "Success"
            }
        except Exception as e:
            return {
                "topics": None,
                "message": f"Error fetching all topics: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")


import asyncio
import struct

import base58
import spl.token.instructions as spl_token
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Confirmed
from solana.rpc.types import TxOpts
#from solana.transaction import Transaction
from solders.transaction import Transaction
from solders.instruction import AccountMeta, Instruction  # type: ignore
from solders.pubkey import Pubkey  # type: ignore
from spl.token.instructions import get_associated_token_address

from agentipy.agent import SolanaAgentKit
from agentipy.constants import (EXPECTED_DISCRIMINATOR, LAMPORTS_PER_SOL,
                                PUMP_EVENT_AUTHORITY, PUMP_FEE, PUMP_GLOBAL,
                                PUMP_PROGRAM,
                                SYSTEM_ASSOCIATED_TOKEN_ACCOUNT_PROGRAM,
                                SYSTEM_PROGRAM, SYSTEM_RENT,
                                SYSTEM_TOKEN_PROGRAM, TOKEN_DECIMALS)
from agentipy.types import BondingCurveState


class PumpfunManager:
    @staticmethod
    async def get_pump_curve_state(conn: AsyncClient, curve_address: Pubkey) -> BondingCurveState:
        response = await conn.get_account_info(curve_address)
        if not response.value or not response.value.data:
            raise ValueError("Invalid curve state: No data")

        data = response.value.data
        if data[:8] != EXPECTED_DISCRIMINATOR:
            raise ValueError("Invalid curve state discriminator")

        return BondingCurveState(data)

    @staticmethod
    def calculate_pump_curve_price(curve_state: BondingCurveState) -> float:
        if curve_state.virtual_token_reserves <= 0 or curve_state.virtual_sol_reserves <= 0:
            raise ValueError("Invalid reserve state")
        return (curve_state.virtual_sol_reserves / LAMPORTS_PER_SOL) / (curve_state.virtual_token_reserves / 10 ** TOKEN_DECIMALS)
    
    @staticmethod
    async def get_token_balance(conn: AsyncClient, associated_token_account: Pubkey):
        response = await conn.get_token_account_balance(associated_token_account)
        if response.value:
            return int(response.value.amount)
        return 0

    @staticmethod
    async def buy_token(agent: SolanaAgentKit, mint: Pubkey, bonding_curve: Pubkey, associated_bonding_curve: Pubkey, amount: float, slippage: float = 0.01, max_retries=5):
        payer = agent.wallet

        async with AsyncClient(agent.rpc_url) as client:
            associated_token_account = get_associated_token_address(payer.pubkey(), mint)
            amount_lamports = int(amount * LAMPORTS_PER_SOL)

            # Fetch the token price
            curve_state = await PumpfunManager.get_pump_curve_state(client, bonding_curve)
            token_price_sol = PumpfunManager.calculate_pump_curve_price(curve_state)
            token_amount = amount / token_price_sol

            # Calculate maximum SOL to spend with slippage
            max_amount_lamports = int(amount_lamports * (1 + slippage))

            # Create associated token account with retries
            for ata_attempt in range(max_retries):
                try:
                    account_info = await client.get_account_info(associated_token_account)
                    if account_info.value is None:
                        print(f"Creating associated token account (Attempt {ata_attempt + 1})...")
                        create_ata_ix = spl_token.create_associated_token_account(
                            payer=payer.pubkey(),
                            owner=payer.pubkey(),
                            mint=mint
                        )
                        create_ata_tx = Transaction()
                        create_ata_tx.add(create_ata_ix)
                        recent_blockhash = await client.get_latest_blockhash()
                        create_ata_tx.recent_blockhash = recent_blockhash.value.blockhash
                        await client.send_transaction(create_ata_tx, payer)
                        print("Associated token account created.")
                        print(f"Associated token account address: {associated_token_account}")
                        break
                    else:
                        print("Associated token account already exists.")
                        print(f"Associated token account address: {associated_token_account}")
                        break
                except Exception as e:
                    print(f"Attempt {ata_attempt + 1} to create associated token account failed: {str(e)}")
                    if ata_attempt < max_retries - 1:
                        wait_time = 2 ** ata_attempt
                        print(f"Retrying in {wait_time} seconds...")
                        await asyncio.sleep(wait_time)
                    else:
                        print("Max retries reached. Unable to create associated token account.")
                        return

            # Continue with the buy transaction
            for attempt in range(max_retries):
                try:
                    accounts = [
                        AccountMeta(pubkey=PUMP_GLOBAL, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=PUMP_FEE, is_signer=False, is_writable=True),
                        AccountMeta(pubkey=mint, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=bonding_curve, is_signer=False, is_writable=True),
                        AccountMeta(pubkey=associated_bonding_curve, is_signer=False, is_writable=True),
                        AccountMeta(pubkey=associated_token_account, is_signer=False, is_writable=True),
                        AccountMeta(pubkey=payer.pubkey(), is_signer=True, is_writable=True),
                        AccountMeta(pubkey=SYSTEM_PROGRAM, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=SYSTEM_TOKEN_PROGRAM, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=SYSTEM_RENT, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=PUMP_EVENT_AUTHORITY, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=PUMP_PROGRAM, is_signer=False, is_writable=False),
                    ]

                    discriminator = struct.pack("<Q", 16927863322537952870)
                    data = discriminator + struct.pack("<Q", int(token_amount * 10**6)) + struct.pack("<Q", max_amount_lamports)
                    buy_ix = Instruction(PUMP_PROGRAM, data, accounts)

                    recent_blockhash = await client.get_latest_blockhash()
                    transaction = Transaction()
                    transaction.add(buy_ix)
                    transaction.recent_blockhash = recent_blockhash.value.blockhash

                    tx = await client.send_transaction(
                        transaction,
                        payer,
                        opts=TxOpts(skip_preflight=True, preflight_commitment=Confirmed),
                    )

                    print(f"Transaction sent: https://explorer.solana.com/tx/{tx.value}")

                    await client.confirm_transaction(tx.value, commitment="confirmed")
                    print("Transaction confirmed")
                    return tx.value

                except Exception as e:
                    print(f"Attempt {attempt + 1} failed: {str(e)}")
                    if attempt < max_retries - 1:
                        wait_time = 2 ** attempt
                        print(f"Retrying in {wait_time} seconds...")
                        await asyncio.sleep(wait_time)
                    else:
                        print("Max retries reached. Unable to complete the transaction.")

    @staticmethod
    async def sell_token(agent: SolanaAgentKit, mint: Pubkey, bonding_curve: Pubkey, associated_bonding_curve: Pubkey, slippage: float = 0.25, max_retries=5):
        payer = agent.wallet

        async with AsyncClient(agent.rpc_url) as client:
            associated_token_account = get_associated_token_address(payer.pubkey(), mint)
            
            # Get token balance
            token_balance = await PumpfunManager.get_token_balance(client, associated_token_account)
            token_balance_decimal = token_balance / 10**TOKEN_DECIMALS
            print(f"Token balance: {token_balance_decimal}")
            if token_balance == 0:
                print("No tokens to sell.")
                return

            # Fetch the token price
            curve_state = await PumpfunManager.get_pump_curve_state(client, bonding_curve)
            token_price_sol = PumpfunManager.calculate_pump_curve_price(curve_state)
            print(f"Price per Token: {token_price_sol:.20f} SOL")

            # Calculate minimum SOL output
            amount = token_balance
            min_sol_output = float(token_balance_decimal) * float(token_price_sol)
            slippage_factor = 1 - slippage
            min_sol_output = int((min_sol_output * slippage_factor) * LAMPORTS_PER_SOL)
            
            print(f"Selling {token_balance_decimal} tokens")
            print(f"Minimum SOL output: {min_sol_output / LAMPORTS_PER_SOL:.10f} SOL")

            for attempt in range(max_retries):
                try:
                    accounts = [
                        AccountMeta(pubkey=PUMP_GLOBAL, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=PUMP_FEE, is_signer=False, is_writable=True),
                        AccountMeta(pubkey=mint, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=bonding_curve, is_signer=False, is_writable=True),
                        AccountMeta(pubkey=associated_bonding_curve, is_signer=False, is_writable=True),
                        AccountMeta(pubkey=associated_token_account, is_signer=False, is_writable=True),
                        AccountMeta(pubkey=payer.pubkey(), is_signer=True, is_writable=True),
                        AccountMeta(pubkey=SYSTEM_PROGRAM, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=SYSTEM_ASSOCIATED_TOKEN_ACCOUNT_PROGRAM, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=SYSTEM_TOKEN_PROGRAM, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=PUMP_EVENT_AUTHORITY, is_signer=False, is_writable=False),
                        AccountMeta(pubkey=PUMP_PROGRAM, is_signer=False, is_writable=False),
                    ]

                    discriminator = struct.pack("<Q", 12502976635542562355)
                    data = discriminator + struct.pack("<Q", amount) + struct.pack("<Q", min_sol_output)
                    sell_ix = Instruction(PUMP_PROGRAM, data, accounts)

                    recent_blockhash = await client.get_latest_blockhash()
                    transaction = Transaction()
                    transaction.add(sell_ix)
                    transaction.recent_blockhash = recent_blockhash.value.blockhash

                    tx = await client.send_transaction(
                        transaction,
                        payer,
                        opts=TxOpts(skip_preflight=True, preflight_commitment=Confirmed),
                    )

                    print(f"Transaction sent: https://explorer.solana.com/tx/{tx.value}")

                    await client.confirm_transaction(tx.value, commitment="confirmed")
                    print("Transaction confirmed")

                    return tx.value

                except Exception as e:
                    print(f"Attempt {attempt + 1} failed: {str(e)}")
                    if attempt < max_retries - 1:
                        wait_time = 2 ** attempt
                        print(f"Retrying in {wait_time} seconds...")
                        await asyncio.sleep(wait_time)
                    else:
                        print("Max retries reached. Unable to complete the transaction.")





from agentipy.agent import SolanaAgentKit
from agentipy.langchain.pumpfun.buy_token import SolanaBuyPumpfunTokenTool
from agentipy.langchain.pumpfun.launch_token import SolanaPumpFunTokenTool
from agentipy.langchain.pumpfun.sell_token import SolanaSellPumpfunTokenTool


def get_pumpfun_tools(solana_kit: SolanaAgentKit):
    return [
        SolanaPumpFunTokenTool(solana_kit=solana_kit),
        SolanaBuyPumpfunTokenTool(solana_kit=solana_kit),
        SolanaSellPumpfunTokenTool(solana_kit=solana_kit)
    ]



from enum import Enum


class ActivationType(Enum):  
    Slot=0
    Timestamp=1

    def __str__(self) -> str:
        return f"{self.value[1]}"
    
    def __repr__(self) -> str:
        return self.name


from agentipy.agent import SolanaAgentKit
from agentipy.langchain.lulo.lend import LuloLendAssetsTool, LuloLendTool
from agentipy.langchain.lulo.withdraw import LuloWithdrawTool


def get_lulo_tools(solana_kit: SolanaAgentKit):
    return [
        LuloLendTool(agent_kit=solana_kit),
        LuloWithdrawTool(agent_kit=solana_kit),
        LuloLendAssetsTool(agent_kit=solana_kit)
    ]


import json

from langchain.tools import BaseTool

from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input


class StorkGetPriceTool(BaseTool):
    name: str = "stork_get_price"
    description: str = """
    Fetch the price of an asset using the Stork Oracle.

    Input: A JSON string with:
    {
        "asset_id": "string, the asset pair ID to fetch price data for (e.g., SOLUSD)."
    }

    Output:
    {
        "price": float, # the token price,
        "timestamp": int, # the unix nanosecond timestamp of the price
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "asset_id": {"type": str, "required": True}
            }
            validate_input(data, schema)
            asset_id = data["asset_id"]
            
            result = await self.solana_kit.stork_fetch_price(asset_id)
            return {
                "status": "success",
                "data": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "message": str(e),
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")



import json

from langchain.tools import BaseTool

from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input


class VirtualsGetSentientListingsTool(BaseTool):
    name: str = "virtuals_get_sentient_listings"
    description: str = """
    Fetches Sentient listings using VirtualsManager.

    Input: A JSON string with:
    {
        "page_number": "int, optional, the page number for pagination (default: 1)",
        "page_size": "int, optional, the number of items per page (default: 30)"
    }
    Output:
    {
        "listings": "dict, the listings data",
        "message": "string, if an error occurs"
    }
    """
    agent_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "page_number": int,
                "page_size": int
            }
            validate_input(data, schema)

            listings = await self.agent_kit.get_sentient_listings(
                page_number=data.get("page_number", 1),
                page_size=data.get("page_size", 30)
            )
            return {
                "listings": listings,
                "message": "Success"
            }
        except Exception as e:
            return {
                "listings": None,
                "message": f"Error fetching Sentient listings: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class VirtualsBuySentientTool(BaseTool):
    name: str = "virtuals_buy_sentient"
    description: str = """
    Purchases Sentient tokens using VirtualsManager.

    Input: A JSON string with:
    {
        "token_address": "string, the token address",
        "amount": "string, the amount to purchase",
        "builder_id": "int, optional, the builder ID for the purchase"
    }
    Output:
    {
        "transaction_receipt": "dict, the transaction receipt",
        "message": "string, if an error occurs"
    }
    """
    agent_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "token_address": str,
                "amount": str,
                "builder_id": int
            }
            validate_input(data, schema)
            
            transaction_receipt = await self.agent_kit.buy_sentient(
                token_address=data["token_address"],
                amount=data["amount"],
                builder_id=data.get("builder_id")
            )
            return {
                "transaction_receipt": transaction_receipt,
                "message": "Success"
            }
        except Exception as e:
            return {
                "transaction_receipt": None,
                "message": f"Error purchasing Sentient tokens: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class VirtualsSellSentientTool(BaseTool):
    name: str = "virtuals_sell_sentient"
    description: str = """
    Sells Sentient tokens using VirtualsManager.

    Input: A JSON string with:
    {
        "token_address": "string, the token address",
        "amount": "string, the amount to sell",
        "builder_id": "int, optional, the builder ID for the sale"
    }
    Output:
    {
        "transaction_receipt": "dict, the transaction receipt",
        "message": "string, if an error occurs"
    }
    """
    agent_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "token_address": str,
                "amount": str,
                "builder_id": int
            }
            validate_input(data, schema)
            
            transaction_receipt = await self.agent_kit.sell_sentient(
                token_address=data["token_address"],
                amount=data["amount"],
                builder_id=data.get("builder_id")
            )
            return {
                "transaction_receipt": transaction_receipt,
                "message": "Success"
            }
        except Exception as e:
            return {
                "transaction_receipt": None,
                "message": f"Error selling Sentient tokens: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")



from agentipy.langchain.allora.allora_get_all_topics import AlloraGetAllTopicsTool
from agentipy.langchain.allora.allora_get_inference_by_topic_id import AlloraGetInferenceByTopicIdTool
from agentipy.langchain.allora.allora_get_price_prediction import AlloraGetPricePredictionTool
from agentipy.agent import SolanaAgentKit

def  get_allora_tools(agent_kit: SolanaAgentKit):
    return [
        AlloraGetAllTopicsTool(agent_kit=agent_kit),
        AlloraGetInferenceByTopicIdTool(agent_kit=agent_kit),
        AlloraGetPricePredictionTool(agent_kit=agent_kit)
    ]


from agentipy.agent import SolanaAgentKit
from agentipy.langchain.squads.approve_multisig_proposal import \
    SquadsApproveMultisigProposalTool
from agentipy.langchain.squads.create_multisig import SquadsCreateMultisigTool
from agentipy.langchain.squads.create_multisig_proposal import \
    SquadsCreateMultisigProposalTool
from agentipy.langchain.squads.deposit_to_multisig_treasury import \
    SquadsDepositToMultisigTreasuryTool
from agentipy.langchain.squads.execute_multisig_proposal import \
    SquadsExecuteMultisigProposalTool
from agentipy.langchain.squads.reject_multisig_proposal import \
    SquadsRejectMultisigProposalTool
from agentipy.langchain.squads.transfer_from_multisig import \
    SquadsTransferFromMultisigTreasuryTool


def get_squads_tools(solana_kit: SolanaAgentKit):
    return [
        SquadsApproveMultisigProposalTool(agent_kit=solana_kit),
        SquadsCreateMultisigTool(agent_kit=solana_kit),
        SquadsCreateMultisigProposalTool(agent_kit=solana_kit),
        SquadsDepositToMultisigTreasuryTool(agent_kit=solana_kit),
        SquadsExecuteMultisigProposalTool(agent_kit=solana_kit),
        SquadsTransferFromMultisigTreasuryTool(agent_kit=solana_kit),
        SquadsRejectMultisigProposalTool(agent_kit=solana_kit),
    ]


# mcp_sse_server.py
import json
import logging
import shlex
from typing import Dict, Union

import uvicorn
from mcp.server import Server
from mcp.server.fastmcp import Context, FastMCP
from mcp.server.sse import SseServerTransport
from mcp.types import TextContent, Tool
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.routing import Mount, Route

logger = logging.getLogger("mcp-sse")

mcp = FastMCP("agentipy-sse-mcp-server")

def parse_key_value_string(s: str) -> dict:
    try:
        return dict(part.split("=", 1) for part in shlex.split(s))
    except Exception as e:
        raise ValueError(f"Invalid key=value string: {e}")

def normalize_kwargs(raw: Union[str, dict]) -> dict:
    if isinstance(raw, dict):
        return raw
    if isinstance(raw, str):
        raw = raw.strip()
        if raw.startswith("{"):
            try:
                raw = json.loads(raw)
                if isinstance(raw, str):
                    raw = json.loads(raw)
                return raw
            except Exception as e:
                raise ValueError(f"Invalid JSON format in 'kwargs': {e}")
        else:
            return parse_key_value_string(raw)
    raise ValueError(f"Unsupported kwargs type: {type(raw)}")

def register_tools(agent, tool_dict: Dict[str, Tool]):
    for name, tool in tool_dict.items():
        def bind_tool(tool_name, tool_def):
            @mcp.tool(name=tool_name, description=tool_def.description)
            async def _tool(ctx: Context, **kwargs):
                try:
                    if "kwargs" in kwargs:
                        kwargs = normalize_kwargs(kwargs["kwargs"])
                    result = await tool_def.handler(agent, kwargs)
                    return TextContent(type="text", text=json.dumps(result, indent=2))
                except Exception as e:
                    logger.error(f"Error in tool {tool_name}: {str(e)}")
                    await ctx.error(f"Error running tool: {str(e)}")
                    return TextContent(type="text", text=f"Error: {str(e)}")
        bind_tool(name, tool)

def create_starlette_app(mcp_server: Server, *, debug: bool = False) -> Starlette:
    sse = SseServerTransport("/messages/")
    async def handle_sse(request: Request) -> None:
        async with sse.connect_sse(request.scope, request.receive, request._send) as (read_stream, write_stream):
            await mcp_server.run(
                read_stream,
                write_stream,
                mcp_server.create_initialization_options(),
            )
    return Starlette(
        debug=debug,
        routes=[
            Route("/sse", endpoint=handle_sse),
            Mount("/messages/", app=sse.handle_post_message),
        ],
    )

def start_sse_server(agent, tools: Dict[str, Tool], host="127.0.0.1", port=3001):
    register_tools(agent, tools)
    app = create_starlette_app(mcp._mcp_server, debug=True)
    logger.info(f"Server running on http://{host}:{port}")
    uvicorn.run(app, host=host, port=port)

def build_sse_starlette_app(agent, tools: Dict[str, Tool], debug=True) -> Starlette:
    register_tools(agent, tools)
    return create_starlette_app(mcp._mcp_server, debug=debug)


import logging
from typing import Any, Dict, Optional

import requests

from agentipy.agent import SolanaAgentKit
from agentipy.utils.agentipy_proxy.utils import encrypt_private_key

logger = logging.getLogger(__name__)

class DeployCollectionManager:
    @staticmethod
    def deploy_collection(
        agent: SolanaAgentKit,
        name: str,
        uri: str,
        royalty_basis_points: int,
        creator_address: str,
    ) -> Optional[Dict[str, Any]]:
        """
        Deploys an NFT collection via an HTTP request to the agent toolkit API.

        :param agent: An instance of SolanaAgentKit configured with required credentials.
        :param name: The name of the NFT collection.
        :param uri: The metadata URI for the collection.
        :param royalty_basis_points: The royalty percentage in basis points (e.g., 500 for 5%).
        :param creator_address: The public key of the collection's creator.
        :return: A dictionary containing the transaction signature or error details.
        """
        try:
            if not all([name, uri, royalty_basis_points, creator_address]):
                raise ValueError("Name, URI, royalty_basis_points, and creator_address are required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "name": name,
                "uri": uri,
                "royaltyBasisPoints": royalty_basis_points,
                "creatorAddress": creator_address,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/nft/deploy-collection",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during collection deployment: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during collection deployment: {error}", exc_info=True)
            return {"success": False, "error": str(error)}
        
    @staticmethod
    def get_metaplex_asset(
        agent: SolanaAgentKit,
        assetId: str
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches detailed information about a specific asset using an HTTP request to the agent toolkit API.

        :param agent: An instance of SolanaAgentKit configured with the required credentials and RPC settings.
        :param assetId: The unique identifier of the asset to be fetched.
        :return: A dictionary containing the asset details or error information.
        """
        try:
            if not all([assetId]):
                raise ValueError("assetId is required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "assetId": assetId,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/nft/get-asset",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during collection deployment: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during collection deployment: {error}", exc_info=True)
            return {"success": False, "error": str(error)}
        
    @staticmethod
    def get_metaplex_assets_by_creator(
        agent: SolanaAgentKit,
        creator: str,
        onlyVerified: bool = False,
        sortBy: Optional[str] = None,
        sortDirection: Optional[str] = None,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches assets created by a specific creator using an HTTP request to the agent toolkit API.

        :param agent: An instance of SolanaAgentKit configured with the required credentials and RPC settings.
        :param creator: The public key of the creator whose assets are to be fetched.
        :param onlyVerified: (Optional) A flag indicating whether to fetch only verified assets (default is False).
        :param sortBy: (Optional) The field by which to sort the assets (e.g., "date", "name").
        :param sortDirection: (Optional) The direction of sorting, either "asc" for ascending or "desc" for descending.
        :param limit: (Optional) The maximum number of assets to retrieve.
        :param page: (Optional) The page number for paginated results.
        :param before: (Optional) Fetch assets listed before this cursor.
        :param after: (Optional) Fetch assets listed after this cursor.
        :return: A dictionary containing the fetched assets or error details.
        """
        try:
            if not all([creator]):
                raise ValueError("creator is required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "creator": creator,
                "onlyVerified": onlyVerified,
                "sortBy": sortBy,
                "sortDirection": sortDirection,
                "limit": limit,
                "page": page,
                "before": before,
                "after": after,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/nft/get-assets-by-creator",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during collection deployment: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during collection deployment: {error}", exc_info=True)
            return {"success": False, "error": str(error)}
        
    @staticmethod
    def get_metaplex_assets_by_authority(
        agent: SolanaAgentKit,
        authority: str,
        sortBy: Optional[str] = None,
        sortDirection: Optional[str] = None,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Fetches assets associated with a specific authority using an HTTP request to the agent toolkit API.

        :param agent: An instance of SolanaAgentKit configured with the required credentials and RPC settings.
        :param authority: The public key of the authority whose assets are to be fetched.
        :param sortBy: (Optional) The field by which to sort the assets (e.g., "date", "name").
        :param sortDirection: (Optional) The direction of sorting, either "asc" for ascending or "desc" for descending.
        :param limit: (Optional) The maximum number of assets to retrieve.
        :param page: (Optional) The page number for paginated results.
        :param before: (Optional) Fetch assets listed before this cursor.
        :param after: (Optional) Fetch assets listed after this cursor.
        :return: A dictionary containing the fetched assets or error details.
        """
        try:
            if not all([authority]):
                raise ValueError("authority is required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "authority": authority,
                "sortBy": sortBy,
                "sortDirection": sortDirection,
                "limit": limit,
                "page": page,
                "before": before,
                "after": after,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/nft/get-assets-by-authority",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during collection deployment: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during collection deployment: {error}", exc_info=True)
            return {"success": False, "error": str(error)}
        
    @staticmethod
    def mint_metaplex_core_nft(
        agent: SolanaAgentKit,
        collectionMint: str,
        name: str,
        uri: str,
        sellerFeeBasisPoints: Optional[int] = None,
        address: Optional[str] = None,
        share: Optional[str] = None,
        recipient: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Mints an NFT using the Metaplex Core program via an HTTP request to the agent toolkit API.

        :param agent: An instance of SolanaAgentKit configured with the required credentials and RPC settings.
        :param collectionMint: The public key of the collection mint to which the NFT belongs.
        :param name: The name of the NFT to be minted.
        :param uri: The metadata URI containing information about the NFT (e.g., image, description).
        :param sellerFeeBasisPoints: (Optional) The royalty percentage in basis points (e.g., 500 for 5% royalties).
        :param address: (Optional) The public key of the creator's address for the NFT.
        :param share: (Optional) The share percentage for the creator (e.g., 100 for 100%).
        :param recipient: (Optional) The public key of the wallet to receive the minted NFT. Defaults to the agent's wallet if not provided.
        :return: A dictionary containing the transaction signature, success status, and any error details.
        """
        try:
            if not all([collectionMint, name, uri]):
                raise ValueError("collectionMint, name, uri is required.")

            encrypted_private_key = encrypt_private_key(agent.private_key)

            payload: Dict[str, Any] = {
                "requestId": encrypted_private_key["requestId"],
                "encrypted_private_key": encrypted_private_key["encryptedPrivateKey"],
                "rpc_url": agent.rpc_url,
                "open_api_key": agent.openai_api_key,
                "collectionMint": collectionMint,
                "name": name,
                "uri": uri,
                "sellerFeeBasisPoints": sellerFeeBasisPoints,
                "address": address,
                "share": share,
                "recipient": recipient,
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/nft/mint",
                json=payload,
                headers={"Content-Type": "application/json"}
            )
            response.raise_for_status()

            data = response.json()
            if data.get("success"):
                return {
                    "success": True,
                    "transaction": data.get("value"),
                    "message": data.get("message"),
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(f"HTTP error during collection deployment: {http_error}", exc_info=True)
            return {"success": False, "error": str(http_error)}
        except ValueError as value_error:
            logger.error(f"Validation error: {value_error}", exc_info=True)
            return {"success": False, "error": str(value_error)}
        except Exception as error:
            logger.error(f"Unexpected error during collection deployment: {error}", exc_info=True)
            return {"success": False, "error": str(error)}


import aiohttp
from agentipy.agent import SolanaAgentKit


class CoingeckoManager:
    @staticmethod
    async def get_trending_tokens(agent: SolanaAgentKit) -> dict:
        """
        Get trending tokens from CoinGecko (Free endpoint).

        Args:
            agent (SolanaAgentKit): The Solana agent instance.

        Returns:
            dict: Trending tokens data.
        """
        try:
            # Always use the free endpoint for trending tokens
            url = "https://api.coingecko.com/api/v3/search/trending"
            # Optionally add a demo key if provided (and no pro key is set)
            if not agent.coingecko_api_key and agent.coingecko_demo_api_key:
                url += f"?x_cg_demo_api_key={agent.coingecko_demo_api_key}"
            
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    if response.status != 200:
                        raise Exception(f"Failed to fetch trending tokens: {response.status}")
                    data = await response.json()
                    return data
        except Exception as e:
            raise Exception(f"Couldn't get trending tokens: {e}")

    @staticmethod
    async def get_trending_pools(agent: SolanaAgentKit, duration: str = "24h") -> dict:
        """
        Get trending pools from CoinGecko for the Solana network (Pro endpoint).

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            duration (str): Duration filter for trending pools. Allowed values: "5m", "1h", "6h", "24h". Default is "24h".

        Returns:
            dict: Trending pools data.
        """
        try:
            if not agent.coingecko_api_key:
                raise Exception("No CoinGecko Pro API key provided")
            
            url = (
                "https://pro-api.coingecko.com/api/v3/onchain/networks/solana/trending_pools"
                f"?include=base_token,network&duration={duration}"
            )
            headers = {"x-cg-pro-api-key": agent.coingecko_api_key}
            
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=headers) as response:
                    if response.status != 200:
                        raise Exception(f"Failed to fetch trending pools: {response.status}")
                    data = await response.json()
                    return data
        except Exception as e:
            raise Exception(f"Error fetching trending pools from CoinGecko: {e}")
        
    @staticmethod
    async def get_top_gainers(
        agent: SolanaAgentKit,
        duration: str = "24h",  # Allowed values: "1h", "24h", "7d", "14d", "30d", "60d", "1y"
        top_coins: int | str = "all"  # Allowed values: 300, 500, 1000, or "all"
    ) -> dict:
        """
        Get top gainers from CoinGecko (Pro endpoint).

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            duration (str): Duration filter for top gainers. Default is "24h".
            top_coins (int or str): The number of top coins to return. Default is "all".

        Returns:
            dict: Top gainers data.
        """
        try:
            if not agent.coingecko_api_key:
                raise Exception("No CoinGecko Pro API key provided")
            
            url = (
                "https://pro-api.coingecko.com/api/v3/coins/top_gainers_losers"
                f"?vs_currency=usd&duration={duration}&top_coins={top_coins}"
            )
            headers = {"x-cg-pro-api-key": agent.coingecko_api_key}
            
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=headers) as response:
                    if response.status != 200:
                        raise Exception(f"Failed to fetch top gainers: {response.status}")
                    data = await response.json()
                    return data
        except Exception as e:
            raise Exception(f"Error fetching top gainers from CoinGecko: {e}")

    @staticmethod
    async def get_token_price_data(agent: SolanaAgentKit, token_addresses: list[str]) -> dict:
        """
        Get token price data for a list of token addresses from CoinGecko (Free endpoint).

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            token_addresses (list[str]): A list of token contract addresses.

        Returns:
            dict: Token price data from CoinGecko.
        """
        try:
            joined_addresses = ",".join(token_addresses)
            url = (
                "https://api.coingecko.com/api/v3/simple/token_price/solana"
                f"?contract_addresses={joined_addresses}"
                "&vs_currencies=usd"
                "&include_market_cap=true"
                "&include_24hr_vol=true"
                "&include_24hr_change=true"
                "&include_last_updated_at=true"
            )
            if not agent.coingecko_api_key and agent.coingecko_demo_api_key:
                url += f"&x_cg_demo_api_key={agent.coingecko_demo_api_key}"
            
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    if response.status != 200:
                        raise Exception(f"Failed to fetch token price data: {response.status}")
                    data = await response.json()
                    return data
        except Exception as e:
            raise Exception(f"Error fetching token price data from CoinGecko: {e}")
        
    @staticmethod
    async def get_token_info(agent: SolanaAgentKit, token_address: str) -> dict:
        """
        Get token info for a given token address from CoinGecko (Pro endpoint).

        Args:
            agent (SolanaAgentKit): The Solana agent instance.
            token_address (str): The token's contract address.

        Returns:
            dict: Token info data.
        """
        try:
            if not agent.coingecko_api_key:
                raise Exception("No CoinGecko Pro API key provided")
            
            url = f"https://pro-api.coingecko.com/api/v3/onchain/networks/solana/tokens/{token_address}/info"
            headers = {"x-cg-pro-api-key": agent.coingecko_api_key}
            
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=headers) as response:
                    if response.status != 200:
                        raise Exception(f"Failed to fetch token info: {response.status}")
                    data = await response.json()
                    return data
        except Exception as e:
            raise Exception(f"Error fetching token info from CoinGecko: {e}")

    @staticmethod
    async def get_latest_pools(agent: SolanaAgentKit) -> dict:
        """
        Get the latest pools from CoinGecko for the Solana network (Pro endpoint).

        Args:
            agent (SolanaAgentKit): The Solana agent instance.

        Returns:
            dict: Latest pools data.
        """
        try:
            if not agent.coingecko_api_key:
                raise Exception("No CoinGecko Pro API key provided")
            
            url = (
                "https://pro-api.coingecko.com/api/v3/onchain/networks/solana/new_pools"
                f"?include=base_token,network"
            )
            headers = {"x-cg-pro-api-key": agent.coingecko_api_key}
            
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=headers) as response:
                    if response.status != 200:
                        raise Exception(f"Failed to fetch latest pools: {response.status}")
                    data = await response.json()
                    return data
        except Exception as e:
            raise Exception(f"Error fetching latest pools from CoinGecko: {e}")
        
    @staticmethod
    async def get_coin_price_vs(agent: SolanaAgentKit, coin_ids: list[str], vs_currencies: list[str] = ["usd"]) -> dict:
        """
        Get trending tokens from CoinGecko (Free endpoint).

        Args:
            agent (SolanaAgentKit): The Solana agent instance.

        Returns:
            dict: Trending tokens data.
        """
        try:
            joined_addresses = ",".join(coin_ids)
            joined_currencies = ",".join(vs_currencies)
       
            url = (
                "https://api.coingecko.com/api/v3/simple/price"
                f"?ids={joined_addresses}"
                f"&vs_currencies={joined_currencies}"
                "&include_market_cap=true"
                "&include_24hr_vol=true"
                "&include_24hr_change=true"
                "&include_last_updated_at=true")
 

            # Optionally add a demo key if provided (and no pro key is set)
            if not agent.coingecko_api_key and agent.coingecko_demo_api_key:
                url += f"?x_cg_demo_api_key={agent.coingecko_demo_api_key}"              
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as response:
                    if response.status != 200:
                        raise Exception(f"Failed to fetch trending tokens: {response.status}")
                    data = await response.json()
                    return data
        except Exception as e:
            raise Exception(f"Couldn't get trending tokens: {e}")



from agentipy.agent import SolanaAgentKit
from agentipy.agent.evm import EvmAgentKit
import requests
import logging
import json
from typing import Union


logger = logging.getLogger(__name__)


class PythManager:
    @staticmethod
    async def get_price(
        agent: Union[SolanaAgentKit, EvmAgentKit], 
        base_token_ticker: str, 
        quote_token_ticker: str
    ):
        """
        Fetch price data for a given token mint address using the Pyth Oracle.
 
        :param agent: Agent kit (Solana or EVM)
        :param base_token_ticker: The ticker of the base token.
        :param quote_token_ticker: The ticker of the quote token.
        :return: A dictionary containing the price, ema_price, and raw_prices.
        """
        try: 
            payload = {
                "base_token_ticker": base_token_ticker,
                "quote_token_ticker": quote_token_ticker
            }

            response = requests.post(
                f"{agent.base_proxy_url}/{agent.api_version}/pyth/getPrice",
                json=payload,
                headers={"Content-Type": "application/json"},
            )

            response.raise_for_status()

            data = response.json()

            if data.get("success"):
                parsed_prices = json.loads(data.get("message"))
                return {
                    "success": True, 
                    "price": int(parsed_prices[0]['price']['price']) * (10 ** parsed_prices[0]['price']['expo']),
                    "ema_price": int(parsed_prices[0]['ema_price']['price']) * (10 ** parsed_prices[0]['ema_price']['expo']),
                    "raw_prices": parsed_prices[0]       
                }
            else:
                return {"success": False, "error": data.get("error", "Unknown error")}

        except requests.exceptions.RequestException as http_error:
            logger.error(
                f"HTTP error during Pyth price retrieval: {http_error}", 
                exc_info=True
            )
            return {"success": False, "error": str(http_error)}
        
        except Exception as error:
            logger.error(
                f"Unexpected error during Pyth price retrieval: {error}", 
                exc_info=True
            )
            return {"success": False, "error": str(error)}


import json
from langchain.tools import BaseTool
from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input

class SolanaGetMetaplexAssetTool(BaseTool):
    name: str = "solana_get_metaplex_asset"
    description: str = """
    Fetches detailed information about a specific Metaplex asset.

    Input: A JSON string with:
    {
        "asset_id": "string, the unique identifier of the asset"
    }

    Output:
    {
        "success": "bool, whether the operation was successful",
        "value": "object, detailed asset information if successful",
        "message": "string, additional details or error information"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "asset_id": {"type": str, "required": True}
            }
            validate_input(data, schema)

            asset_id = data["asset_id"]

            result = await self.solana_kit.get_metaplex_asset(asset_id)
            return result
        except Exception as e:
            return {"success": False, "message": f"Error fetching Metaplex asset: {str(e)}"}

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")



from agentipy.agent import SolanaAgentKit
from agentipy.langchain.tensor.trade import (TensorCancelListingTool,
                                             TensorListNFTForSaleTool)


def get_tensor_tools(solana_kit: SolanaAgentKit):
    return [
        TensorCancelListingTool(solana_kit),
        TensorListNFTForSaleTool(solana_kit),
    ]


import json

from langchain.tools import BaseTool

from agentipy.agent import SolanaAgentKit
from agentipy.helpers import validate_input


class TensorListNFTForSaleTool(BaseTool):
    name: str = "tensor_list_nft_for_sale"
    description: str = """
    Lists an NFT for sale on Tensor.

    Input: A JSON string with:
    {
        "price": "float, the sale price of the NFT",
        "nft_mint": "string, the NFT mint address"
    }

    Output:
    {
        "transaction_details": "dict, transaction details",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "price": {"type": float, "required": True},
                "nft_mint": {"type": str, "required": True},
            }
            validate_input(data, schema)

            price = data["price"]
            nft_mint = data["nft_mint"]

            result = await self.solana_kit.list_nft_for_sale(price, nft_mint)
            return {
                "status": "success",
                "transaction_details": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "message": str(e),
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")


class TensorCancelListingTool(BaseTool):
    name: str = "tensor_cancel_listing"
    description: str = """
    Cancels an NFT listing on Tensor.

    Input: A JSON string with:
    {
        "nft_mint": "string, the NFT mint address"
    }

    Output:
    {
        "transaction_details": "dict, transaction details",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            schema = {
                "nft_mint": {"type": str, "required": True},
            }
            validate_input(data, schema)

            nft_mint = data["nft_mint"]

            result = await self.solana_kit.cancel_listing(nft_mint)
            return {
                "status": "success",
                "transaction_details": result,
            }
        except Exception as e:
            return {
                "status": "error",
                "message": str(e),
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")



from agentipy.agent import SolanaAgentKit
from langchain.tools import BaseTool
from agentipy.helpers import validate_input
import json

class SolanaFetchPriceTool(BaseTool):
    """
    Tool to fetch the price of a token in USDC.
    """
    name:str = "solana_fetch_price"
    description:str = """Fetch the price of a given token in USDC.

    Inputs:
    - tokenId: string, the mint address of the token, e.g., "JUPyiwrYJFskUPiHa7hkeR8VUtAeFoSYbKedZNsDvCN"
    """
    solana_kit: SolanaAgentKit

    async def call(self, input: str) -> str:
        try:
            data = json.loads(input)
            schema = {
                "token_id": {"type": str, "required": True}
            }
            validate_input(data, schema)
            token_id = data["token_id"]
            price = await self.solana_kit.fetch_price(token_id)
            return json.dumps({
                "status": "success",
                "tokenId": token_id,
                "priceInUSDC": price,
            })
        except Exception as error:
            return json.dumps({
                "status": "error",
                "message": str(error),
                "code": getattr(error, "code", "UNKNOWN_ERROR"),
            })
        
    def _run(self):
        """Synchronous version of the run method, required by BaseTool."""
        raise NotImplementedError(
            "This tool only supports async execution via _arun. Please use the async interface."
        )



from agentipy.agent import SolanaAgentKit
from agentipy.langchain.moonshot.buy import SolanaBuyUsingMoonshotTool
from agentipy.langchain.moonshot.sell import SolanaSellUsingMoonshotTool


def get_moonshot_tools(solana_kit: SolanaAgentKit):
    return [
        SolanaSellUsingMoonshotTool(solana_kit=solana_kit),
        SolanaBuyUsingMoonshotTool(solana_kit=solana_kit)
    ]





import json
from agentipy.agent import SolanaAgentKit
from langchain.tools import BaseTool
from agentipy.helpers import validate_input

class BackpackGetStatusTool(BaseTool):
    name: str = "backpack_get_status"
    description: str = """
    Fetches the system status and any status messages using the BackpackManager.

    Input: None
    Output:
    {
        "status": "dict, the system status",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self):
        try:
            status = await self.solana_kit.get_status()
            return {
                "status": status,
                "message": "Success"
            }
        except Exception as e:
            return {
                "status": None,
                "message": f"Error fetching system status: {str(e)}"
            }

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackSendPingTool(BaseTool):
    name: str = "backpack_send_ping"
    description: str = """
    Sends a ping and expects a "pong" response using the BackpackManager.

    Input: None
    Output:
    {
        "response": "string, the response ('pong')",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self):
        try:
            response = await self.solana_kit.send_ping()
            return {
                "response": response,
                "message": "Success"
            }
        except Exception as e:
            return {
                "response": None,
                "message": f"Error sending ping: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetSystemTimeTool(BaseTool):
    name: str = "backpack_get_system_time"
    description: str = """
    Fetches the current system time using the BackpackManager.

    Input: None
    Output:
    {
        "system_time": "string, the current system time",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self):
        try:
            system_time = await self.solana_kit.get_system_time()
            return {
                "system_time": system_time,
                "message": "Success"
            }
        except Exception as e:
            return {
                "system_time": None,
                "message": f"Error fetching system time: {str(e)}"
            }

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetSupportedAssetsTool(BaseTool):
    name: str = "backpack_get_supported_assets"
    description: str = """
    Fetches supported assets using the BackpackManager.

    Input: None
    Output:
    {
        "assets": "list, the supported assets",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self):
        try:
            assets = await self.solana_kit.get_supported_assets()
            return {
                "assets": assets,
                "message": "Success"
            }
        except Exception as e:
            return {
                "assets": None,
                "message": f"Error fetching supported assets: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetTickerInformationTool(BaseTool):
    name: str = "backpack_get_ticker_information"
    description: str = """
    Fetches ticker information using the BackpackManager.

    Input: A JSON string with optional filters for the query.
    Output:
    {
        "ticker_information": "dict, the ticker information",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            data = json.loads(input)
            ticker_info = await self.solana_kit.get_ticker_information(**data)
            return {
                "ticker_information": ticker_info,
                "message": "Success"
            }
        except Exception as e:
            return {
                "ticker_information": None,
                "message": f"Error fetching ticker information: {str(e)}"
            }

    def _run(self, input: str):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")

class BackpackGetCollateralInfoTool(BaseTool):
    name: str = "backpack_get_collateral_info"
    description: str = """
    Fetches collateral information using the BackpackManager.

    Input: A JSON string with:
    {
        "sub_account_id": "int, optional, the sub-account ID for collateral information"
    }
    Output:
    {
        "collateral_info": "dict, the collateral information",
        "message": "string, if an error occurs"
    }
    """
    solana_kit: SolanaAgentKit

    async def _arun(self, input: str):
        try:
            schema = {
                "sub_account_id": {"type": int, "required": False}
            }
            data = json.loads(input)
            validate_input(data, schema)

            collateral_info = await self.solana_kit.get_collateral_info(
                sub_account_id=data.get("sub_account_id")
            )
            return {
                "collateral_info": collateral_info,
                "message": "Success"
            }
        except Exception as e:
            return {
                "collateral_info": None,
                "message": f"Error fetching collateral information: {str(e)}"
            }

    def _run(self):
        raise NotImplementedError("This tool only supports async execution via _arun. Please use the async interface.")
